import sys
import numpy
from pyscf.pbc.dft import numint as pyscf_numint
from pyscf.pbc.dft.gen_grid import make_mask, BLKSIZE
from pyscfad.lib import numpy as np
from pyscfad.lib import ops, stop_grad
from pyscfad.dft import numint
from pyscfad.dft.numint import eval_mat, _contract_rho, _dot_ao_dm

def nr_rks(ni, cell, grids, xc_code, dms, spin=0, relativity=0, hermi=0,
           kpts=None, kpts_band=None, max_memory=2000, verbose=None):
    if kpts is None:
        kpts = numpy.zeros((1,3))

    xctype = ni._xc_type(xc_code)
    make_rho, nset, nao = ni._gen_rho_evaluator(cell, dms, hermi)

    nelec = [0]*nset
    excsum = [0]*nset
    vmat = [0]*nset
    if xctype == 'LDA':
        ao_deriv = 0
        for ao_k1, ao_k2, mask, weight, coords \
                in ni.block_loop(cell, grids, nao, ao_deriv, kpts, kpts_band,
                                 max_memory):
            for i in range(nset):
                rho = make_rho(i, ao_k2, mask, xctype)
                exc, vxc = ni.eval_xc(xc_code, rho, spin=0,
                                      relativity=relativity, deriv=1)[:2]
                den = rho*weight
                nelec[i] += stop_grad(den).sum()
                excsum[i] += (den*exc).sum()
                vmat[i] += ni.eval_mat(cell, ao_k1, weight, rho, vxc,
                                       mask, xctype, 0, verbose)
    elif xctype == 'GGA':
        ao_deriv = 1
        for ao_k1, ao_k2, mask, weight, coords \
                in ni.block_loop(cell, grids, nao, ao_deriv, kpts, kpts_band,
                                 max_memory):
            for i in range(nset):
                rho = make_rho(i, ao_k2, mask, xctype)
                exc, vxc = ni.eval_xc(xc_code, rho, spin=0,
                                      relativity=relativity, deriv=1)[:2]
                den = rho[0]*weight
                nelec[i] += stop_grad(den).sum()
                excsum[i] += (den*exc).sum()
                vmat[i] += ni.eval_mat(cell, ao_k1, weight, rho, vxc,
                                       mask, xctype, 0, verbose)
    elif xctype == 'MGGA':
        if any(x in xc_code.upper() for x in ('CC06', 'CS', 'BR89', 'MK00')):
            raise NotImplementedError('laplacian in meta-GGA method')
        ao_deriv = 2
        for ao_k1, ao_k2, mask, weight, coords \
                in ni.block_loop(cell, grids, nao, ao_deriv, kpts, kpts_band,
                                 max_memory):
            for i in range(nset):
                rho = make_rho(i, ao_k2, mask, xctype)
                exc, vxc = ni.eval_xc(xc_code, rho, spin=0,
                                      relativity=relativity, deriv=1)[:2]
                den = rho[0]*weight
                nelec[i] += stop_grad(den).sum()
                excsum[i] += (den*exc).sum()
                vmat[i] += ni.eval_mat(cell, ao_k1, weight, rho, vxc,
                                       mask, xctype, 0, verbose)

    nelec = numpy.asarray(nelec)
    excsum = np.asarray(excsum)
    vmat = np.asarray(vmat)
    if nset == 1:
        nelec = nelec[0]
        excsum = excsum[0]
        vmat = vmat[0]
    return nelec, excsum, vmat

def eval_ao(cell, coords, kpt=numpy.zeros(3), deriv=0, relativity=0, shls_slice=None,
            non0tab=None, out=None, verbose=None):
    ao_kpts = eval_ao_kpts(cell, coords, numpy.reshape(kpt, (-1,3)), deriv,
                           relativity, shls_slice, non0tab, out, verbose)
    return ao_kpts[0]

def eval_ao_kpts(cell, coords, kpts=None, deriv=0, relativity=0,
                 shls_slice=None, non0tab=None, out=None, verbose=None, **kwargs):
    if kpts is None:
        if 'kpt' in kwargs:
            sys.stderr.write('WARN: KNumInt.eval_ao function finds keyword '
                             'argument "kpt" and converts it to "kpts"\n')
            kpts = kwargs['kpt']
        else:
            kpts = numpy.zeros((1,3))
    kpts = numpy.reshape(kpts, (-1,3))

    comp = (deriv+1)*(deriv+2)*(deriv+3)//6
    if cell.cart:
        feval = 'GTOval_cart_deriv%d' % deriv
    else:
        feval = 'GTOval_sph_deriv%d' % deriv
    return cell.pbc_eval_gto(feval, coords, comp, kpts,
                             shls_slice=shls_slice, non0tab=non0tab, out=out)

def eval_rho(cell, ao, dm, non0tab=None, xctype='LDA', hermi=0, verbose=None):
    if xctype in ['LDA', 'HF']:
        ngrids, nao = ao.shape
    else:
        ngrids, nao = ao[0].shape

    if non0tab is None:
        non0tab = numpy.empty(((ngrids+BLKSIZE-1)//BLKSIZE, cell.nbas),
                              dtype=numpy.uint8)
        non0tab[:] = 0xff

    # complex orbitals or density matrix
    if np.iscomplexobj(ao) or np.iscomplexobj(dm):
        shls_slice = (0, cell.nbas)
        ao_loc = cell.ao_loc_nr()
        dm = np.asarray(dm, dtype=np.complex128)

# For GGA, function eval_rho returns   real(|\nabla i> D_ij <j| + |i> D_ij <\nabla j|)
#       = real(|\nabla i> D_ij <j| + |i> D_ij <\nabla j|)
#       = real(|\nabla i> D_ij <j| + conj(|\nabla j> conj(D_ij) < i|))
#       = real(|\nabla i> D_ij <j|) + real(|\nabla j> conj(D_ij) < i|)
#       = real(|\nabla i> [D_ij + (D^\dagger)_ij] <j|)
# symmetrization dm (D + D.conj().T) then /2 because the code below computes
#       2*real(|\nabla i> D_ij <j|)
        if not hermi:
            dm = (dm + dm.conj().T) * .5

        def dot_bra(bra, aodm):
            # rho = numpy.einsum('pi,pi->p', bra.conj(), aodm).real
            #:rho  = numpy.einsum('pi,pi->p', bra.real, aodm.real)
            #:rho += numpy.einsum('pi,pi->p', bra.imag, aodm.imag)
            #:return rho
            return _contract_rho(bra, aodm)

        if xctype == 'LDA' or xctype == 'HF':
            c0 = _dot_ao_dm(cell, ao, dm, non0tab, shls_slice, ao_loc)
            rho = dot_bra(ao, c0)
        elif xctype == 'GGA':
            rho = np.empty((4,ngrids))
            c0 = _dot_ao_dm(cell, ao[0], dm, non0tab, shls_slice, ao_loc)
            #rho[0] = dot_bra(ao[0], c0)
            rho = ops.index_update(rho, ops.index[0], dot_bra(ao[0], c0))
            for i in range(1, 4):
                #rho[i] = dot_bra(ao[i], c0) * 2
                rho = ops.index_update(rho, ops.index[i], dot_bra(ao[i], c0) * 2)
        else:
            # rho[4] = \nabla^2 rho, rho[5] = 1/2 |nabla f|^2
            rho = np.empty((6,ngrids))
            c0 = _dot_ao_dm(cell, ao[0], dm, non0tab, shls_slice, ao_loc)
            #rho[0] = dot_bra(ao[0], c0)
            rho = ops.index_update(rho, ops.index[0], dot_bra(ao[0], c0))
            #rho[5] = 0
            rho = ops.index_update(rho, ops.index[5], 0)
            for i in range(1, 4):
                #rho[i] = dot_bra(ao[i], c0) * 2  # *2 for +c.c.
                rho = ops.index_update(rho, ops.index[i], dot_bra(ao[i], c0) * 2)
                c1 = _dot_ao_dm(cell, ao[i], dm, non0tab, shls_slice, ao_loc)
                #rho[5] += dot_bra(ao[i], c1)
                rho = ops.index_add(rho, ops.index[5], dot_bra(ao[i], c1))
            XX, YY, ZZ = 4, 7, 9
            ao2 = ao[XX] + ao[YY] + ao[ZZ]
            #rho[4] = dot_bra(ao2, c0)
            rho = ops.index_update(rho, ops.index[4], dot_bra(ao2, c0))
            #rho[4] += rho[5]
            rho = ops.index_add(rho, ops.index[4], rho[5])
            #rho[4] *= 2 # *2 for +c.c.
            rho = ops.index_mul(rho, ops.index[4], 2)
            #rho[5] *= .5
            rho = ops.index_mul(rho, ops.index[5], .5)
    else:
        # real orbitals and real DM
        rho = numint.eval_rho(cell, ao, dm, non0tab, xctype, hermi, verbose)
    return rho

class NumInt(numint.NumInt):
    def nr_rks(self, cell, grids, xc_code, dms, hermi=0,
               kpt=numpy.zeros(3), kpts_band=None, max_memory=2000, verbose=None):
        if kpts_band is not None:
            # To compute Vxc on kpts_band, convert the NumInt object to KNumInt object.
            ni = KNumInt()
            ni.__dict__.update(self.__dict__)
            nao = dms.shape[-1]
            return ni.nr_rks(cell, grids, xc_code, dms.reshape(-1,1,nao,nao),
                             hermi, kpt.reshape(1,3), kpts_band, max_memory,
                             verbose)
        return nr_rks(self, cell, grids, xc_code, dms,
                      0, 0, hermi, kpt, kpts_band, max_memory, verbose)

    def eval_ao(self, cell, coords, kpt=numpy.zeros(3), deriv=0, relativity=0,
                shls_slice=None, non0tab=None, out=None, verbose=None):
        return eval_ao(cell, coords, kpt, deriv, relativity, shls_slice,
                       non0tab, out, verbose)

    def eval_mat(self, cell, ao, weight, rho, vxc,
                 non0tab=None, xctype='LDA', spin=0, verbose=None):
        # Guess whether ao is evaluated for kpts_band.  When xctype is LDA, ao on grids
        # should be a 2D array.  For other xc functional, ao should be a 3D array.
        if ao.ndim == 2 or (xctype != 'LDA' and ao.ndim == 3):
            mat = eval_mat(cell, ao, weight, rho, vxc, non0tab, xctype, spin, verbose)
        else:
            nkpts = len(ao)
            nao = ao[0].shape[-1]
            mat = np.empty((nkpts,nao,nao), dtype=np.complex128)
            for k in range(nkpts):
                mat[k] = eval_mat(cell, ao[k], weight, rho, vxc,
                                  non0tab, xctype, spin, verbose)
        return mat

    def block_loop(self, cell, grids, nao=None, deriv=0, kpt=numpy.zeros(3),
                   kpts_band=None, max_memory=2000, non0tab=None, blksize=None):
        '''Define this macro to loop over grids by blocks.
        '''
        # For UniformGrids, grids.coords does not indicate whehter grids are initialized
        if grids.non0tab is None:
            grids.build(with_non0tab=True)
        if nao is None:
            nao = cell.nao
        grids_coords = grids.coords
        grids_weights = grids.weights
        ngrids = grids_coords.shape[0]
        comp = (deriv+1)*(deriv+2)*(deriv+3)//6
# NOTE to index grids.non0tab, the blksize needs to be the integer multiplier of BLKSIZE
        if blksize is None:
            blksize = int(max_memory*1e6/(comp*2*nao*16*BLKSIZE))*BLKSIZE
            blksize = max(BLKSIZE, min(blksize, ngrids, BLKSIZE*1200))
        if non0tab is None:
            non0tab = grids.non0tab
        if non0tab is None:
            non0tab = numpy.empty(((ngrids+BLKSIZE-1)//BLKSIZE,cell.nbas),
                                  dtype=numpy.uint8)
            non0tab[:] = 0xff
        kpt = numpy.reshape(kpt, 3)
        if kpts_band is None:
            kpt1 = kpt2 = kpt
        else:
            kpt1 = kpts_band
            kpt2 = kpt

        for ip0 in range(0, ngrids, blksize):
            ip1 = min(ngrids, ip0+blksize)
            coords = grids_coords[ip0:ip1]
            weight = grids_weights[ip0:ip1]
            non0 = non0tab[ip0//BLKSIZE:]
            ao_k2 = self.eval_ao(cell, coords, kpt2, deriv=deriv, non0tab=non0)
            if abs(kpt1-kpt2).sum() < 1e-9:
                ao_k1 = ao_k2
            else:
                ao_k1 = self.eval_ao(cell, coords, kpt1, deriv=deriv)
            yield ao_k1, ao_k2, non0, weight, coords
            ao_k1 = ao_k2 = None

class KNumInt(numint.NumInt):
    def __init__(self, kpts=numpy.zeros((1,3))):
        numint.NumInt.__init__(self)
        self.kpts = kpts #numpy.reshape(kpts, (-1,3))

    def nr_rks(self, cell, grids, xc_code, dms, hermi=0, kpts=None, kpts_band=None,
               max_memory=2000, verbose=None, **kwargs):
        if kpts is None:
            if 'kpt' in kwargs:
                sys.stderr.write('WARN: KNumInt.nr_rks function finds keyword '
                                 'argument "kpt" and converts it to "kpts"\n')
                kpts = kwargs['kpt']
            else:
                kpts = self.kpts
        kpts = kpts.reshape(-1,3)

        return nr_rks(self, cell, grids, xc_code, dms, 0, 0,
                      hermi, kpts, kpts_band, max_memory, verbose)

    def eval_mat(self, cell, ao_kpts, weight, rho, vxc,
                 non0tab=None, xctype='LDA', spin=0, verbose=None):
        nkpts = len(ao_kpts)
        nao = ao_kpts[0].shape[-1]
        #dtype = np.result_type(*ao_kpts)
        #mat = np.empty((nkpts,nao,nao), dtype=dtype)
        mat = [0] * nkpts
        for k in range(nkpts):
            mat[k] = eval_mat(cell, ao_kpts[k], weight, rho, vxc,
                              non0tab, xctype, spin, verbose)
        return np.asarray(mat)

    def eval_ao(self, cell, coords, kpts=numpy.zeros((1,3)), deriv=0, relativity=0,
                shls_slice=None, non0tab=None, out=None, verbose=None, **kwargs):
        return eval_ao_kpts(cell, coords, kpts, deriv,
                            relativity, shls_slice, non0tab, out, verbose)

    def eval_rho(self, cell, ao_kpts, dm_kpts, non0tab=None, xctype='LDA',
                 hermi=0, verbose=None):
        nkpts = len(ao_kpts)
        rhoR = 0
        for k in range(nkpts):
            rhoR += eval_rho(cell, ao_kpts[k], dm_kpts[k], non0tab, xctype,
                             hermi, verbose)
        rhoR *= 1./nkpts
        return rhoR

    def _gen_rho_evaluator(self, cell, dms, hermi=0):
        if getattr(dms, 'mo_coeff', None) is not None:
            raise NotImplementedError
        else:
            if getattr(dms[0], "ndim", 0) == 2:
                dms = [np.stack(dms)]
            nao = dms[0].shape[-1]
            ndms = len(dms)

            def make_rho(idm, ao_kpts, non0tab, xctype):
                return self.eval_rho(cell, ao_kpts, dms[idm], non0tab, xctype,
                                     hermi=hermi)
        return make_rho, ndms, nao

    block_loop = pyscf_numint.KNumInt.block_loop



from pyscf.prop.polarizability.rhf import Polarizability as pyscf_Polarizability
from pyscfad import util

Traced_Attributes = ['_scf', 'mol']

@util.pytree_node(Traced_Attributes, num_args=1)
class Polarizability(pyscf_Polarizability):
    def __init__(self, mf, **kwargs):
        pyscf_Polarizability.__init__(self, mf)
        self.__dict__.update(kwargs)



import os
import numpy

def load_library(libname):
    try:
        _loaderpath = os.path.dirname(__file__)
        return numpy.ctypeslib.load_library(libname, _loaderpath)
    except OSError:
        print ("Can't load library %s" % libname)
        raise



from functools import partial
import numpy
from jax import custom_jvp, custom_vjp
from pyscf.pbc.gto import cell
from pyscfad.lib import numpy as jnp

@partial(custom_jvp, nondiff_argnums=tuple(range(1,7)))
def _pbc_intor(mol, intor, comp=None, hermi=0, kpts=None, kpt=None,
               shls_slice=None):
    return cell.Cell.pbc_intor(mol, intor, comp, hermi, kpts, kpt, shls_slice)

@_pbc_intor.defjvp
def _pbc_intor_jvp(intor, comp, hermi, kpts, kpt, shls_slice,
                   primals, tangents):
    mol, = primals
    mol_t, = tangents

    primal_out = _pbc_intor(mol, intor, comp, hermi, kpts, kpt, shls_slice)
    tangent_out = None
    if mol.coords is not None:
        if intor.startswith("int1e"):
            intor_ip = intor.replace("int1e_", "int1e_ip")
        else:
            raise NotImplementedError
        tangent_out = _int1e_jvp_r0(mol, mol_t, intor_ip, kpts, kpt, shls_slice)

    if mol.ctr_coeff is not None:
        pass
    if mol.exp is not None:
        pass

    return primal_out, tangent_out

def _int1e_jvp_r0(mol, mol_t, intor, kpts, kpt, shls_slice):
    coords_t = mol_t.coords
    aoslices = mol.aoslice_by_atom()
    nao = mol.nao
    s1 = cell.Cell.pbc_intor(mol, intor, comp=3, hermi=0, 
                             kpts=kpts, kpt=kpt, shls_slice=shls_slice)

    gamma = False
    if isinstance(s1, numpy.ndarray):
        s1 = [s1,]
        gamma = True

    def get_tangent_out(s1_k):
        grad = numpy.zeros((mol.natm,3,nao,nao), dtype=s1_k.dtype)
        for ia in range(mol.natm):
            p0, p1 = aoslices [ia,2:]
            grad[ia,:,p0:p1] += -s1_k[:,p0:p1]
        tangent_out_k = jnp.einsum("nxij,nx->ij", grad, coords_t)
        tangent_out_k += tangent_out_k.T.conj()
        return tangent_out_k

    # FIXME this may be slow with many k-points
    tangent_out = [get_tangent_out(s1_k) for s1_k in s1]

    if gamma:
        tangent_out = tangent_out[0]
    return tangent_out


@partial(custom_vjp, nondiff_argnums=tuple(range(1,7)))
def _pbc_intor_rev(mol, intor, comp=None, hermi=0, kpts=None, kpt=None,
                   shls_slice=None):
    return cell.Cell.pbc_intor(mol, intor, comp, hermi, kpts, kpt, shls_slice)

def _pbc_intor_fwd(mol, intor, comp, hermi, kpts, kpt, shls_slice):
    primal_out = _pbc_intor_rev(mol, intor, comp, hermi, kpts, kpt, shls_slice)
    res = (mol,)
    return primal_out, res

def _pbc_intor_bwd(intor, comp, hermi, kpts, kpt, shls_slice, res, y_bar):
    mol, = res

    if mol.coords is not None:
        if intor.startswith("int1e"):
            intor_ip = intor.replace("int1e_", "int1e_ip")
        else:
            raise NotImplementedError
    partial_r0 = _int1e_partial_r0(mol, intor_ip, kpts, kpt, shls_slice)
    if isinstance(partial_r0, numpy.ndarray):
        r0_bar = jnp.einsum("nxij,ij->nx", partial_r0, y_bar)
    else:
        r0_bar = []
        for item in partial_r0:
            r0_bar.append(jnp.einsum("nxij,ij->nx", item, y_bar))

    ctr_coeff_bar=None
    if mol.ctr_coeff is not None:
        pass

    exp_bar=None
    if mol.exp is not None:
       pass

    mol.coords = r0_bar
    return (mol,)

def _int1e_partial_r0(mol, intor, kpts, kpt, shls_slice):
    aoslices = mol.aoslice_by_atom()
    nao = mol.nao
    s1 = cell.Cell.pbc_intor(mol, intor, comp=3, hermi=0,
                             kpts=kpts, kpt=kpt, shls_slice=shls_slice)

    if isinstance(s1, numpy.ndarray):
        s1 = [s1,]
    nkpts = len(s1)

    partial = []
    for k in range(nkpts):
        s1_k = s1[k]
        grad = numpy.zeros((mol.natm,3,nao,nao), dtype=s1_k.dtype)
        for ia in range(mol.natm):
            p0, p1 = aoslices [ia,2:]
            grad[ia,:,p0:p1] += -s1_k[:,p0:p1]
            grad[ia,:,:,p0:p1] += -s1_k[:,p0:p1].transpose(0,2,1).conj()
        #grad += grad.transpose(0,1,3,2)
        partial.append(grad)
    if nkpts == 1:
        partial = partial[0]
    return partial

_pbc_intor_rev.defvjp(_pbc_intor_fwd, _pbc_intor_bwd)



import numpy
from pyscf.hessian import thermo
from pyscf.data.nist import BOHR

def harmonic_analysis(mol, hess, ir_tensor=None, raman_tensor=None,
                      exclude_trans=True, exclude_rot=True,
                      imaginary_freq=True):
    '''
    Computes harmonic vibrational frequency, normal modes,
    IR intensity, Raman activity, Depolarization ratio.

    Args:
        hess: (natm,3,natm,3) array
            The 2nd order derivative of energy w.r.t. atomic coordinates
        ir_tensor: (3,natm,3) array
            The derivative of dipole moment w.r.t. atomic coordinates
        raman_tensor: (3,3,natm,3) array
            The derivative of polarizability w.r.t. atomic coordinates
        exclude_trans: bool
            Whether to exclude translation modes
        exclude_rot: bool
            Whether to exclude rotatoin modes
        imaginary_freq: bool
            Whether to allow imaginary frequency

    Returns:
        vibration: dict
            Results for vibrational analysis
        ir: dict
            Results for IR
        raman: dict
            Results for Raman
    '''

    vibration = thermo.harmonic_analysis(mol, hess.transpose(0,2,1,3),
                                   exclude_trans, exclude_rot, imaginary_freq)
    ir = compute_ir(ir_tensor, vibration)
    raman = compute_raman(raman_tensor, vibration)
    return vibration, ir, raman

def compute_ir(ir_tensor, vibration):
    ir = {'intensity': None} # km/mol
    if ir_tensor is None:
        return ir

    unit_kmmol = 974.8801118351438
    norm_mode = vibration['norm_mode']
    a = numpy.einsum('inx,knx->ki', ir_tensor, norm_mode)
    ir['intensity'] = numpy.einsum('ki,ki->k', a, a) * unit_kmmol
    return ir

def compute_raman(raman_tensor, vibration):
    raman = {'activity': None, # A^4/amu
             'depolar_ratio': None}
    if raman_tensor is None:
        return raman

    norm_mode = vibration['norm_mode']
    chi = raman_tensor * BOHR**2
    alpha = numpy.einsum('ijnx,knx->kij', chi, norm_mode)
    alpha2 = (1./3. * numpy.trace(alpha, axis1=1, axis2=2)) ** 2

    alpha_diag = numpy.diagonal(alpha, axis1=1, axis2=2)
    alpha_ij = alpha_diag[:,:,None] - alpha_diag[:,None,:]
    gamma2  = numpy.einsum('kij->k', 1./4. * alpha_ij ** 2)
    gamma2 += numpy.einsum('kij->k', 1.5 * alpha ** 2)
    gamma2 -= numpy.einsum('kii->k', 1.5 * alpha ** 2)

    raman['activity'] = 45 * alpha2 + 7 * gamma2
    raman['depolar_ratio'] = 3 * gamma2 / (45 * alpha2 + 4 * gamma2)
    return raman



import numpy
import scipy.linalg
from pyscf import __config__
from pyscf.lib import logger
from pyscf.lib import diis as pyscf_diis
from pyscfad.lib import numpy as np

class DIIS(pyscf_diis.DIIS):
    def __init__(self, dev=None, filename=None,
                 incore=getattr(__config__, 'lib_diis_DIIS_incore', False)):
        pyscf_diis.DIIS.__init__(self, dev=dev, filename=filename, incore=incore)

    def extrapolate(self, nd=None):
        if nd is None:
            nd = self.get_num_vec()
        if nd == 0:
            raise ValueError('No vector found in DIIS object.')

        h = self._H[:nd+1,:nd+1]
        g = numpy.zeros(nd+1, h.dtype)
        g[0] = 1

        w, v = scipy.linalg.eigh(h)
        if numpy.any(abs(w)<1e-14):
            logger.debug(self, 'Linear dependence found in DIIS error vectors.')
            idx = abs(w)>1e-14
            c = numpy.dot(v[:,idx]*(1./w[idx]), numpy.dot(v[:,idx].T.conj(), g))
        else:
            try:
                c = numpy.linalg.solve(h, g)
            except numpy.linalg.linalg.LinAlgError as e:
                logger.warn(self, ' diis singular, eigh(h) %s', w)
                raise e
        logger.debug1(self, 'diis-c %s', c)

        xnew = None
        for i, ci in enumerate(c[1:]):
            xi = self.get_vec(i)
            if xnew is None:
                xnew = np.zeros(xi.size, c.dtype)
            xnew += xi * ci
        return xnew



from typing import Optional, Any
from pyscf import __config__
from pyscf.cc import eom_rccsd as pyscf_eom_rccsd
from pyscfad import lib
from pyscfad.lib import numpy as np
from pyscfad import gto
from pyscfad.cc import ccsd


@lib.dataclass
class EOM(pyscf_eom_rccsd.EOM):
    _cc: ccsd.CCSD = lib.field(pytree_node=True)
    mol: Optional[gto.Mole] = None
    verbose: Optional[int] = None
    stdout: Any = None
    max_memory: Optional[int] = None
    max_space: int = getattr(__config__, 'eom_rccsd_EOM_max_space', 20)
    max_cycle: Optional[int] = None
    conv_tol: Optional[float] = None
    partition: Any = getattr(__config__, 'eom_rccsd_EOM_partition', None)
    
    e: Optional[float] = None
    v: Optional[np.array] = None
    nocc: Optional[np.array] = None
    nmo: Optional[np.array] = None
   
    def __post_init__(self):
        if self.mol is None:
            self.mol = self._cc.mol
        if self.verbose is None:
            self.verbose = self._cc.verbose
        if self.stdout is None:
            self.stdout = self._cc.stdout
        if self.max_memory is None:
            self.max_memory = self._cc.max_memory
        if self.max_cycle is None:
            self.max_cycle = getattr(__config__, 'eom_rccsd_EOM_max_cycle', self._cc.max_cycle)
        if self.conv_tol is None:
            self.conv_tol = getattr(__config__, 'eom_rccsd_EOM_conv_tol', self._cc.conv_tol)

        if self.nocc is None:
            self.nocc = self._cc.nocc
        if self.nmo is None:
            self.nmo = self._cc.nmo 
        
        self._keys = set(self.__dict__.keys())

def ipccsd_diag(eom, imds=None):
    if imds is None: imds = eom.make_imds()
    t1, t2 = imds.t1, imds.t2
    dtype = np.result_type(t1, t2)
    nocc, nvir = t1.shape
    fock = imds.eris.fock
    foo = fock[:nocc,:nocc]
    fvv = fock[nocc:,nocc:]

    Hr1 = -np.diag(imds.Loo)
    Hr2 = np.zeros((nocc,nocc,nvir), dtype)

    if eom.partition == 'mp':
        foo_diag = np.diag(foo)
        fvv_diag = np.diag(fvv)
        Hr2 = - foo_diag[:,None,None] - foo_diag[None,:,None]  + fvv_diag[None,None,:]
    else:
        Lvv_diag = np.diag(imds.Lvv)
        Loo_diag = np.diag(imds.Loo)
        Hr2 = - Loo_diag[:,None,None] - Loo_diag[None,:,None]  + Lvv_diag[None,None,:]
        wij = np.einsum('ijij->ij', imds.Woooo)
        Hr2 += wij[:,:,None]
        wjb = np.einsum('jbbj->jb', imds.Wovvo)
        Hr2 += 2 * wjb[None,:,:]
        Hr2 += -np.einsum('ij,jb->ijb', np.eye(nocc), wjb)
        wjb = np.einsum('jbjb->jb', imds.Wovov)
        Hr2 += -wjb[None,:,:]
        Hr2 += -wjb[:,None,:]
        Hr2 += -2*np.einsum('jibc,ijcb->ijb', imds.Woovv, t2)
        Hr2 += np.einsum('ijbc,ijcb->ijb', imds.Woovv, t2)

    vector = eom.amplitudes_to_vector(Hr1, Hr2)
    return vector

def ipccsd_matvec(eom, vector, imds=None, diag=None):
    # Ref: Nooijen and Snijders, J. Chem. Phys. 102, 1681 (1995) Eqs.(8)-(9)
    if imds is None: imds = eom.make_imds()
    nocc = eom.nocc
    nmo = eom.nmo
    r1, r2 = eom.vector_to_amplitudes(vector, nmo, nocc)

    # 1h-1h block
    Hr1 = -np.einsum('ki,k->i', imds.Loo, r1)
    #1h-2h1p block
    Hr1 += 2*np.einsum('ld,ild->i', imds.Fov, r2)
    Hr1 +=  -np.einsum('kd,kid->i', imds.Fov, r2)
    Hr1 += -2*np.einsum('klid,kld->i', imds.Wooov, r2)
    Hr1 +=    np.einsum('lkid,kld->i', imds.Wooov, r2)

    # 2h1p-1h block
    Hr2 = -np.einsum('kbij,k->ijb', imds.Wovoo, r1)
    # 2h1p-2h1p block
    if eom.partition == 'mp':
        fock = imds.eris.fock
        foo = fock[:nocc,:nocc]
        fvv = fock[nocc:,nocc:]
        Hr2 += np.einsum('bd,ijd->ijb', fvv, r2)
        Hr2 += -np.einsum('ki,kjb->ijb', foo, r2)
        Hr2 += -np.einsum('lj,ilb->ijb', foo, r2)
    elif eom.partition == 'full':
        diag_matrix2 = vector_to_amplitudes_ip(diag, nmo, nocc)[1]
        Hr2 += diag_matrix2 * r2
    else:
        Hr2 += np.einsum('bd,ijd->ijb', imds.Lvv, r2)
        Hr2 += -np.einsum('ki,kjb->ijb', imds.Loo, r2)
        Hr2 += -np.einsum('lj,ilb->ijb', imds.Loo, r2)
        Hr2 +=  np.einsum('klij,klb->ijb', imds.Woooo, r2)
        Hr2 += 2*np.einsum('lbdj,ild->ijb', imds.Wovvo, r2)
        Hr2 +=  -np.einsum('kbdj,kid->ijb', imds.Wovvo, r2)
        Hr2 +=  -np.einsum('lbjd,ild->ijb', imds.Wovov, r2) #typo in Ref
        Hr2 +=  -np.einsum('kbid,kjd->ijb', imds.Wovov, r2)
        tmp = 2*np.einsum('lkdc,kld->c', imds.Woovv, r2)
        tmp += -np.einsum('kldc,kld->c', imds.Woovv, r2)
        Hr2 += -np.einsum('c,ijcb->ijb', tmp, imds.t2)

    vector = eom.amplitudes_to_vector(Hr1, Hr2)
    return vector

class EOMIP(EOM, pyscf_eom_rccsd.EOMIP):
    get_diag = ipccsd_diag
    matvec = ipccsd_matvec

class EOMEA(EOM, pyscf_eom_rccsd.EOMEA):
    pass

class EOMEE(EOM, pyscf_eom_rccsd.EOMEE):
    pass



from pyscfad.lib import numpy as np

def get_jk(dfobj, dm, hermi=1, with_j=True, with_k=True, direct_scf_tol=1e-13):
    assert(with_j or with_k)

    nao = dfobj.mol.nao
    dms = dm.reshape(-1, nao, nao)

    vj = vk = 0
    Lpq = dfobj._cderi
    if with_j:
        tmp = np.einsum("Lpq,xpq->xL", Lpq, dms)
        vj = np.einsum("Lpq,xL->xpq", Lpq, tmp)
        vj = vj.reshape(dm.shape)

    if with_k:
        tmp = np.einsum('Lij,xjk->xLki', Lpq, dms)
        vk = np.einsum('Lki,xLkj->xij', Lpq, tmp)
        vk = vk.reshape(dm.shape)
    return vj, vk



from pyscf import numpy as np
from pyscf.scf import uhf as pyscf_uhf
from pyscfad import util
from pyscfad.scf import hf

@util.pytree_node(hf.Traced_Attributes, num_args=1)
class UHF(hf.SCF, pyscf_uhf.UHF):
    def __init__(self, mol, **kwargs):
        pyscf_uhf.UHF.__init__(self, mol)
        self.__dict__.update(kwargs)

    def eig(self, h, s, x0=None):
        if x0 is not None:
            e_a, c_a = self._eigh(h[0], s, x0[0])
            e_b, c_b = self._eigh(h[1], s, x0[1])
        else:
            e_a, c_a = self._eigh(h[0], s)
            e_b, c_b = self._eigh(h[1], s)
        return np.array((e_a,e_b)), np.array((c_a,c_b))



import sys
import numpy
from pyscf import __config__
from pyscf import numpy as np
from pyscf.lib import logger
from pyscf.pbc.scf import hf as pyscf_pbc_hf
from pyscf.pbc.scf.hf import _format_jks
from pyscfad import lib
from pyscfad import util
from pyscfad.lib import stop_grad
from pyscfad.scf import hf as mol_hf
from pyscfad.pbc import df

Traced_Attributes = ['cell', 'mo_coeff', 'mo_energy', 'with_df']

@util.pytree_node(Traced_Attributes, num_args=1)
class SCF(mol_hf.SCF, pyscf_pbc_hf.SCF):
    def __init__(self, cell, kpt=numpy.zeros(3),
                 exxdiv=getattr(__config__, 'pbc_scf_SCF_exxdiv', 'ewald'), **kwargs):
        if not cell._built:
            sys.stderr.write('Warning: cell.build() is not called in input\n')
            cell.build()
        self.cell = cell
        mol_hf.SCF.__init__(self, cell)

        self.with_df = df.FFTDF(cell)
        # Range separation JK builder
        self.rsjk = None

        self.exxdiv = exxdiv
        self.kpt = kpt
        self.conv_tol = cell.precision * 10
        self.no_incore = True

        self._keys = self._keys.union(['cell', 'exxdiv', 'with_df'])
        self.__dict__.update(kwargs)

    def get_init_guess(self, cell=None, key='minao'):
        if cell is None: cell = self.cell
        dm = mol_hf.SCF.get_init_guess(self, cell, key)
        dm = normalize_dm_(self, dm)
        return dm

    def get_hcore(self, cell=None, kpt=None):
        if cell is None: cell = self.cell
        if kpt is None: kpt = self.kpt
        if cell.pseudo:
            nuc = self.with_df.get_pp(kpt, cell=cell)
        else:
            raise NotImplementedError
            #nuc = self.with_df.get_nuc(kpt)
        if len(cell._ecpbas) > 0:
            raise NotImplementedError
            #nuc += ecp.ecp_int(cell, kpt)
        return nuc + cell.pbc_intor('int1e_kin', 1, 1, kpt)

    def get_jk(self, cell=None, dm=None, hermi=1, kpt=None, kpts_band=None,
               with_j=True, with_k=True, omega=None, **kwargs):
        if cell is None: cell = self.cell
        if dm is None: dm = self.make_rdm1()
        if kpt is None: kpt = self.kpt

        #cpu0 = (logger.process_clock(), logger.perf_counter())
        dm = np.asarray(dm)
        nao = dm.shape[-1]

        if (not omega and kpts_band is None and
            # TODO: generate AO integrals with rsjk algorithm
            not self.rsjk and
            (self.exxdiv == 'ewald' or not self.exxdiv) and
            (self._eri is not None or cell.incore_anyway or
             (not self.direct_scf and self._is_mem_enough())) and not self.no_incore):
            raise NotImplementedError
            #if self._eri is None:
            #    logger.debug(self, 'Building PBC AO integrals incore')
            #    self._eri = self.with_df.get_ao_eri(kpt, compact=True)
            #vj, vk = mol_hf.dot_eri_dm(self._eri, dm, hermi, with_j, with_k)

            #if with_k and self.exxdiv == 'ewald':
            #    from pyscf.pbc.df.df_jk import _ewald_exxdiv_for_G0
            #    # G=0 is not inculded in the ._eri integrals
            #    _ewald_exxdiv_for_G0(self.cell, kpt, dm.reshape(-1,nao,nao),
            #                         vk.reshape(-1,nao,nao))
        elif self.rsjk:
            raise NotImplementedError
            #vj, vk = self.rsjk.get_jk(dm.reshape(-1,nao,nao), hermi, kpt, kpts_band,
            #                          with_j, with_k, omega, exxdiv=self.exxdiv)
        else:
            vj, vk = self.with_df.get_jk(dm.reshape(-1,nao,nao), hermi, kpt, kpts_band,
                                         with_j, with_k, omega, exxdiv=self.exxdiv, cell=cell)

        if with_j:
            vj = _format_jks(vj, dm, kpts_band)
        if with_k:
            vk = _format_jks(vk, dm, kpts_band)
        #logger.timer(self, 'vj and vk', *cpu0)
        return vj, vk

    get_veff = pyscf_pbc_hf.SCF.get_veff
    energy_nuc = pyscf_pbc_hf.SCF.energy_nuc
    energy_grad = mol_hf.SCF.energy_grad

RHF = SCF

def normalize_dm_(mf, dm):
    cell = mf.cell
    s = stop_grad(mf.get_ovlp(cell))
    if getattr(dm, "ndim", 0) == 2:
        ne = numpy.einsum('ij,ji->', stop_grad(dm), s).real
    else:
        ne = numpy.einsum('xij,ji->', stop_grad(dm), s).real
    if abs(ne - cell.nelectron).sum() > 1e-7:
        logger.debug(mf, 'Big error detected in the electron number '
                     'of initial guess density matrix (Ne/cell = %g)!\n'
                     '  This can cause huge error in Fock matrix and '
                     'lead to instability in SCF for low-dimensional '
                     'systems.\n  DM is normalized wrt the number '
                     'of electrons %s', ne, cell.nelectron)
        dm *= cell.nelectron / ne
    return dm



import sys
import numpy
from pyscf import __config__
from pyscf import numpy as np
from pyscf.pbc.scf import khf as pyscf_khf
from pyscfad import lib
from pyscfad import util
from pyscfad.lib import stop_grad
from pyscfad.scf import hf as mol_hf
from pyscfad.pbc import df
from pyscfad.pbc.scf import hf as pbchf

# TODO add mo_coeff, which requires AD wrt complex numbers
Traced_Attributes = ['cell', 'mo_energy', 'with_df']

def get_hcore(mf, cell=None, kpts=None):
    if cell is None: cell = mf.cell
    if kpts is None: kpts = mf.kpts
    if cell.pseudo:
        nuc = mf.with_df.get_pp(kpts, cell=cell)
    else:
        raise NotImplementedError
        #nuc = lib.asarray(mf.with_df.get_nuc(kpts))
    if len(cell._ecpbas) > 0:
        raise NotImplementedError
        #nuc += lib.asarray(ecp.ecp_int(cell, kpts))
    t = np.asarray(cell.pbc_intor('int1e_kin', 1, 1, kpts))
    return nuc + t


@util.pytree_node(Traced_Attributes, num_args=1)
class KSCF(pbchf.SCF, pyscf_khf.KSCF):
    def __init__(self, cell, kpts=numpy.zeros((1,3)),
                 exxdiv=getattr(__config__, 'pbc_scf_SCF_exxdiv', 'ewald'), **kwargs):
        if not cell._built:
            sys.stderr.write('Warning: cell.build() is not called in input\n')
            cell.build()
        self.cell = cell
        mol_hf.SCF.__init__(self, cell)

        self.with_df = df.FFTDF(cell, kpts=kpts)
        # Range separation JK builder
        self.rsjk = None

        self.exxdiv = exxdiv
        #self.kpts = kpts
        self.conv_tol = cell.precision * 10

        self.exx_built = False
        self._keys = self._keys.union(['cell', 'exx_built', 'exxdiv', 'with_df', 'rsjk'])
        self.__dict__.update(kwargs)

    def get_jk(self, cell=None, dm_kpts=None, hermi=1, kpts=None, kpts_band=None,
               with_j=True, with_k=True, omega=None, **kwargs):
        if cell is None: cell = self.cell
        if kpts is None: kpts = self.kpts
        if dm_kpts is None: dm_kpts = self.make_rdm1()
        #cpu0 = (logger.process_clock(), logger.perf_counter())
        if self.rsjk:
            raise NotImplementedError
            #vj, vk = self.rsjk.get_jk(dm_kpts, hermi, kpts, kpts_band,
            #                          with_j, with_k, omega, self.exxdiv)
        else:
            vj, vk = self.with_df.get_jk(dm_kpts, hermi, kpts, kpts_band,
                                         with_j, with_k, omega, self.exxdiv, cell=cell)
        #logger.timer(self, 'vj and vk', *cpu0)
        return vj, vk

    def get_init_guess(self, cell=None, key='minao'):
        if cell is None:
            cell = self.cell
        cell = stop_grad(cell)
        return pyscf_khf.KSCF.get_init_guess(self, cell, key)

    def eig(self, h_kpts, s_kpts, x0=None):
        nkpts = len(h_kpts)
        eig_kpts = []
        mo_coeff_kpts = []
        if x0 is None:
            x0 = [None] * nkpts

        for k in range(nkpts):
            e, c = self._eigh(h_kpts[k], s_kpts[k], x0[k])
            eig_kpts.append(e)
            mo_coeff_kpts.append(c)
        return eig_kpts, mo_coeff_kpts


    get_hcore = get_hcore
    get_ovlp = pyscf_khf.KSCF.get_ovlp
    get_fock = pyscf_khf.KSCF.get_fock
    get_occ = pyscf_khf.KSCF.get_occ
    energy_elec = pyscf_khf.KSCF.energy_elec
    get_fermi = pyscf_khf.KSCF.get_fermi

    get_veff = pyscf_khf.KSCF.get_veff
    get_j = pyscf_khf.KSCF.get_j
    get_k = pyscf_khf.KSCF.get_k
    get_grad = pyscf_khf.KSCF.get_grad
    make_rdm1 = pyscf_khf.KSCF.make_rdm1
    #eig = pyscf_khf.KSCF.eig

KRHF = KSCF



from functools import partial
from jax import jit, custom_jvp
from pyscf.dft import libxc
from pyscf.dft.libxc import parse_xc, is_lda, is_meta_gga
from pyscfad.lib import numpy as np

def eval_xc(xc_code, rho, spin=0, relativity=0, deriv=1, omega=None, verbose=None):
    # NOTE only consider exc and vxc
    if deriv > 1:
        raise NotImplementedError

    hyb, fn_facs = parse_xc(xc_code)
    if omega is not None:
        hyb[2] = float(omega)

    exc = _eval_xc_comp(rho, hyb, fn_facs, spin, relativity, deriv=0, verbose=verbose)
    if deriv == 0:
        vxc = (None,) * 4
    else:
        vxc = _eval_xc_comp(rho, hyb, fn_facs, spin, relativity, deriv=1, verbose=verbose)
    return exc, vxc, None, None

@partial(custom_jvp, nondiff_argnums=tuple(range(1,7)))
def _eval_xc_comp(rho, hyb, fn_facs, spin=0, relativity=0, deriv=1, verbose=None):
    return libxc._eval_xc(hyb, fn_facs, rho, spin, relativity, deriv, verbose)[deriv]

@_eval_xc_comp.defjvp
def _eval_xc_comp_jvp(hyb, fn_facs, spin, relativity, deriv, verbose,
                      primals, tangents):
    rho, = primals
    rho_t, = tangents
    if deriv > 2:
        raise NotImplementedError

    val  = _eval_xc_comp(rho, hyb, fn_facs, spin, relativity, deriv, verbose)
    val1 = _eval_xc_comp(rho, hyb, fn_facs, spin, relativity, deriv+1, verbose)

    fn_ids = [x[0] for x in fn_facs]
    n = len(fn_ids)
    if (n == 0 or
        all((is_lda(x) for x in fn_ids))):
        if deriv == 0:
            jvp = (val1[0] - val) / rho * rho_t
        elif deriv == 1:
            jvp = (val1[0] * rho_t,) + (None,) * 3
        else:
            jvp = (val1[0] * rho_t,) + (None,) * 9
    elif any((is_meta_gga(x) for x in fn_ids)):
        if deriv == 0:
            exc1 = _exc_partial_deriv(rho, val, val1, "MGGA")
            jvp = np.einsum('np,np->p', exc1, rho_t)
        elif deriv == 1:
            vrho1, vsigma1, vlapl1, vtau1 = _vxc_partial_deriv(rho, val, val1, "MGGA")
            vrho_jvp = np.einsum('np,np->p', vrho1, rho_t)
            vsigma_jvp = np.einsum('np,np->p', vsigma1, rho_t)
            vlapl_jvp = np.einsum('np,np->p', vlapl1, rho_t)
            vtau_jvp = np.einsum('np,np->p', vtau1, rho_t)
            vrho1 = vsigma1 = vlapl1 = vtau1 = None
            jvp = np.vstack((vrho_jvp, vsigma_jvp, vlapl_jvp, vtau_jvp))
        else:
            raise NotImplementedError
    else:
        if deriv == 0:
            exc1 = _exc_partial_deriv(rho, val, val1, "GGA")
            jvp = np.einsum('np,np->p', exc1, rho_t)
        elif deriv == 1:
            vrho1, vsigma1 = _vxc_partial_deriv(rho, val, val1, "GGA")[:2]
            vrho_jvp = np.einsum('np,np->p', vrho1, rho_t)
            vsigma_jvp = np.einsum('np,np->p', vsigma1, rho_t)
            vrho1 = vsigma1 = None
            jvp = (vrho_jvp, vsigma_jvp, None, None)
        else:
            v2rho2, v2rhosigma, v2sigma2 = _fxc_partial_deriv(rho, val, val1, "GGA")[:3]
            v2rho2_jvp = np.einsum('np,np->p', v2rho2, rho_t)
            v2rhosigma_jvp = np.einsum('np,np->p', v2rhosigma, rho_t)
            v2sigma2_jvp = np.einsum('np,np->p', v2sigma2, rho_t)
            jvp = (v2rho2_jvp, v2rhosigma_jvp, v2sigma2_jvp) + (None,)*7
    return val, jvp

@partial(jit, static_argnames=['xctype'])
def _exc_partial_deriv(rho, exc, vxc, xctype="LDA"):
    if xctype == "LDA":
        exc1 = (vxc[0] - exc) / rho
    elif xctype in ["GGA", "MGGA"]:
        drho = (vxc[0] - exc) / rho[0]
        dsigma = vxc[1] / rho[0] * 2. * rho[1:4]
        exc1 = np.vstack((drho, dsigma))
        if xctype == "MGGA":
            dlap = vxc[2] / rho[0]
            dtau = vxc[3] / rho[0]
            exc1 = np.vstack((exc1, dlap, dtau))
    else:
        raise KeyError
    return exc1

@partial(jit, static_argnames=['xctype'])
def _vxc_partial_deriv(rho, vxc, fxc, xctype="LDA"):
    vrho1 = vsigma1 = vlapl1 = vtau1 = None
    if xctype == "LDA":
        vrho1 = fxc[0]
    elif xctype in ["GGA", "MGGA"]:
        vrho1 = np.vstack((fxc[0], fxc[1] * 2. * rho[1:4]))
        vsigma1 = np.vstack((fxc[1], fxc[2] * 2. * rho[1:4]))
        if xctype == "MGGA":
            vrho1 = np.vstack((vrho1, fxc[5], fxc[6]))
            vsigma1 = np.vstack((vsigma1, fxc[8], fxc[9]))
            vlapl1 = np.vstack((fxc[5], fxc[8] * 2. * rho[1:4], fxc[3], fxc[7]))
            vtau1 = np.vstack((fxc[6], fxc[9] * 2. * rho[1:4], fxc[7], fxc[4]))
    else:
        raise KeyError
    return vrho1, vsigma1, vlapl1, vtau1

@partial(jit, static_argnames=['xctype'])
def _fxc_partial_deriv(rho, fxc, kxc, xctype="LDA"):
    v2rho2_1 = v2rhosigma_1 = v2sigma2_1 = None
    if xctype == "LDA":
        v2rho2_1 = kxc[0]
    elif xctype in ["GGA", "MGGA"]:
        v2rho2_1 = np.vstack((kxc[0], kxc[1] * 2. * rho[1:4]))
        v2rhosigma_1 = np.vstack((kxc[1], kxc[2] * 2. * rho[1:4]))
        v2sigma2_1 = np.vstack((kxc[2], kxc[3] * 2. * rho[1:4]))
        if xctype == "MGGA":
            raise NotImplementedError
    else:
        raise KeyError
    return (v2rho2_1, v2rhosigma_1, v2sigma2_1) + (None,) * 7



import warnings
from functools import partial
import numpy
#from scipy.special import erfc
from jax import custom_jvp
from jax.scipy.special import erfc
from pyscf import __config__
from pyscf.gto.mole import PTR_COORD
from pyscf.gto.moleintor import _get_intor_and_comp
from pyscf.pbc.gto import cell as pyscf_cell
from pyscfad import lib
from pyscfad import util
from pyscfad.lib import numpy as np
from pyscfad.lib import stop_grad
from pyscfad.gto import mole
from pyscfad.gto._mole_helper import setup_exp, setup_ctr_coeff
from pyscfad.pbc.gto import _pbcintor
from pyscfad.pbc.gto.eval_gto import eval_gto as pbc_eval_gto

def get_SI(cell, Gv=None):
    coords = cell.atom_coords()
    ngrids = numpy.prod(cell.mesh)
    if Gv is None or Gv.shape[0] == ngrids:
        Gv = cell.get_Gv()
    GvT = Gv.T
    SI = np.exp(-1j*np.dot(coords, GvT))
    return SI

energy_nuc = pyscf_cell.ewald

def shift_bas_center(cell0, r):
    cell = cell0.copy()
    cell.coords = cell0.atom_coords() + r[None,:]

    ptr = cell._atm[:,PTR_COORD]
    idx = numpy.vstack((ptr, ptr+1, ptr+2)).T.flatten()
    numpy.put(cell._env, idx, stop_grad(cell.coords).flatten())
    return cell

def intor_cross(intor, cell1, cell2, comp=None, hermi=0, kpts=None, kpt=None,
                shls_slice=None, **kwargs):
    intor, comp = _get_intor_and_comp(cell1._add_suffix(intor), comp)

    if kpts is None:
        if kpt is not None:
            kpts_lst = np.reshape(kpt, (1,3))
        else:
            kpts_lst = np.zeros((1,3))
    else:
        kpts_lst = np.reshape(kpts, (-1,3))

    Ls = cell2.get_lattice_Ls(rcut=max(cell1.rcut, cell2.rcut))
    expkL = np.exp(1j*np.dot(kpts_lst, Ls.T))

    ni = cell1.nao
    nj = cell2.nao
    nL = len(Ls)
    ints = []
    for i in range(nL):
        shifted_cell = shift_bas_center(cell2, Ls[i])
        ints.append(mole.intor_cross(intor, cell1, shifted_cell, comp=comp))
    ints = np.asarray(ints)

    if comp == 1:
        out = np.einsum('kl,lij->kij', expkL, ints)
    else:
        out = np.einsum('kl,lcij->kcij', expkL, ints)

    if kpts is None or np.shape(kpts) == (3,):  # A single k-point
        out = out[0]
    return out

def pbc_intor(cell, intor, comp=None, hermi=0, kpts=None, kpt=None,
              shls_slice=None, **kwargs):
    if kwargs:
        warnings.warn("Keyword arguments %s are ignored" % list(kwargs.keys()))

    if cell.abc is None:
        res = _pbcintor._pbc_intor(cell, intor, comp=comp, hermi=hermi, kpts=kpts,
                                   kpt=kpt, shls_slice=shls_slice)
    else:
        res = intor_cross(intor, cell, cell, comp=comp, hermi=hermi, kpts=kpts,
                          kpt=kpt, shls_slice=shls_slice, **kwargs)
    return res

@util.pytree_node(mole.Traced_Attributes+['abc'])
class Cell(mole.Mole, pyscf_cell.Cell):
    def __init__(self, **kwargs):
        mole.Mole.__init__(self, **kwargs)
        self.a = None # lattice vectors, (a1,a2,a3)
        self.abc = None # traced lattice vectors
        # if set, defines a spherical cutoff
        # of fourier components, with .5 * G**2 < ke_cutoff
        self.ke_cutoff = None

        self.pseudo = None
        self.dimension = 3
        # TODO: Simple hack for now; the implementation of ewald depends on the
        #       density-fitting class.  This determines how the ewald produces
        #       its energy.
        self.low_dim_ft_type = None

##################################################
# These attributes are initialized by build function if not given
        self.mesh = None
        self.rcut = None

##################################################
# don't modify the following variables, they are not input arguments
        keys = ('precision', 'exp_to_discard')
        self._keys = self._keys.union(self.__dict__).union(keys)
        self.__dict__.update(kwargs)


    def build(self, *args, **kwargs):
        trace_coords = kwargs.pop("trace_coords", True)
        trace_exp = kwargs.pop("trace_exp", False)
        trace_ctr_coeff = kwargs.pop("trace_ctr_coeff", False)
        trace_r0 = kwargs.pop("trace_r0", False)
        trace_lattice_vectors = kwargs.pop("trace_lattice_vectors", False)

        pyscf_cell.Cell.build(self, *args, **kwargs)

        if trace_coords:
            self.coords = np.asarray(self.atom_coords())
        if trace_exp:
            self.exp, _, _ = setup_exp(self)
        if trace_ctr_coeff:
            self.ctr_coeff, _, _ = setup_ctr_coeff(self)
        if trace_r0:
            pass
        if trace_lattice_vectors:
            self.abc = np.asarray(self.lattice_vectors())

    def lattice_vectors(self):
        if self.abc is None:
            return pyscf_cell.Cell.lattice_vectors(self)
        else:
            return self.abc

    def pbc_eval_gto(self, eval_name, coords, comp=None, kpts=None, kpt=None,
                     shls_slice=None, non0tab=None, ao_loc=None, out=None):
        return pbc_eval_gto(self, eval_name, coords, comp, kpts, kpt,
                            shls_slice, non0tab, ao_loc, out)
    pbc_eval_ao = pbc_eval_gto

    def eval_gto(self, eval_name, coords, comp=None, kpts=None, kpt=None,
                 shls_slice=None, non0tab=None, ao_loc=None, out=None):
        if eval_name[:3] == 'PBC':
            return self.pbc_eval_gto(eval_name, coords, comp, kpts, kpt,
                                     shls_slice, non0tab, ao_loc, out)
        else:
            return mole.eval_gto(self, eval_name, coords, comp,
                                 shls_slice, non0tab, ao_loc, out)
    eval_ao = eval_gto
    pbc_intor = pbc_intor
    get_SI = get_SI
    energy_nuc = energy_nuc



from jax import scipy
from pyscf import numpy as np
from pyscfad.lib.numpy_helper import unpack_triu

def update_rotate_matrix(dx, u0=1):
    dr = unpack_triu(dx, filltril=2)
    u = np.dot(u0, scipy.linalg.expm(dr))
    return u

def rotate_mo(mo_coeff, u):
    mo = np.dot(mo_coeff, u)
    return mo

def rotate_mo1(mo_coeff, x):
    u = update_rotate_matrix(x)
    mo_coeff1 = rotate_mo(mo_coeff, u)
    return mo_coeff1

def rotate_mo1_ov(mo_coeff, x, nocc):
    u = update_rotate_matrix(x)
    u = u.at[:nocc,:nocc].set(0.)
    u = u.at[nocc:,nocc:].set(0.)
    mo_coeff1 = rotate_mo(mo_coeff, u)
    return mo_coeff1



from functools import partial
import numpy
from jax import scipy
from jax import custom_jvp
from pyscf import lib
from pyscf.lib import logger
from pyscf import gto
from pyscf.gto.moleintor import getints
from pyscf.df.outcore import _guess_shell_ranges
from pyscf import __config__
from pyscfad.lib import ops
from pyscfad.lib import numpy as np
from . import addons

MAX_MEMORY = getattr(__config__, 'df_outcore_max_memory', 2000)
LINEAR_DEP_THR = getattr(__config__, 'df_df_DF_lindep', 1e-12)

@partial(custom_jvp, nondiff_argnums=tuple(range(2,7)))
def int3c_cross(mol, auxmol, intor="int3c2e", comp=1, aosym="s1", out=None, shls_slice=None):
    assert aosym == 's1'
    int3c = gto.moleintor.ascint3(mol._add_suffix(intor))
    nbas = mol.nbas
    nauxbas = auxmol.nbas
    if shls_slice is None:
        shls_slice = (0, nbas, 0, nbas, nbas, nbas+nauxbas)
    pmol = mol + auxmol
    ints = gto.moleintor.getints3c(int3c, pmol._atm, pmol._bas, pmol._env,
                                   shls_slice=shls_slice, comp=comp,
                                   aosym=aosym, ao_loc=None, cintopt=None, out=out)
    return ints

@int3c_cross.defjvp
def int3c_cross_jvp(intor, comp, aosym, out, shls_slice, primals, tangents):
    mol, auxmol = primals
    mol_dot, auxmol_dot = tangents

    primal_out = int3c_cross(mol, auxmol, intor=intor, comp=comp,
                             aosym=aosym, out=out, shls_slice=shls_slice)
    tangent_out = np.zeros_like(primal_out)

    if intor.startswith("int3c2e") and not "spinor" in intor:
        intor_ip1 = intor.replace("int3c2e", "int3c2e_ip1")
        ints = int3c_cross(mol, auxmol, intor=intor_ip1, comp=3,
                           aosym=aosym, out=None, shls_slice=shls_slice)
        tangent_out_mol = _int3c_fill_grad_r0_ip1(mol, mol_dot, -ints)

        intor_ip2 = intor.replace("int3c2e", "int3c2e_ip2")
        ints = int3c_cross(mol, auxmol, intor=intor_ip2, comp=3,
                           aosym=aosym, out=None, shls_slice=shls_slice)
        tangent_out_auxmol = _int3c_fill_grad_r0_ip2(auxmol, auxmol_dot, -ints)
        tangent_out += tangent_out_mol + tangent_out_auxmol
    else:
        raise NotImplementedError

    if mol.ctr_coeff is not None:
        print("ctr_coeff derivative for int3c2e not implemented")
    if mol.exp is not None:
        print("exp derivative for int3c2e not implemented")

    return primal_out, tangent_out

def _int3c_fill_grad_r0_ip1(mol, mol_dot, ints):
    shape = [mol.natm,] + list(ints.shape)
    grad = np.zeros(shape, dtype=ints.dtype)
    aoslices = mol.aoslice_by_atom()
    for ia in range(mol.natm):
        p0, p1 = aoslices[ia,2:]
        grad = ops.index_update(grad, ops.index[ia,:,p0:p1], ints[:,p0:p1])
    tangent_out = np.einsum('nxijk,nx->ijk', grad, mol_dot.coords)
    tangent_out += tangent_out.transpose(1,0,2)
    return tangent_out

def _int3c_fill_grad_r0_ip2(mol, mol_dot, ints):
    shape = [mol.natm,] + list(ints.shape)
    grad = np.zeros(shape, dtype=ints.dtype)
    aoslices = mol.aoslice_by_atom()
    for ia in range(mol.natm):
        p0, p1 = aoslices[ia,2:]
        grad = ops.index_update(grad, ops.index[ia,...,p0:p1], ints[...,p0:p1])
    tangent_out = np.einsum('nxijk,nx->ijk', grad, mol_dot.coords)
    return tangent_out


def cholesky_eri(mol, auxmol=None, auxbasis='weigend+etb',
                 int3c='int3c2e', aosym='s1', int2c='int2c2e', comp=1,
                 max_memory=MAX_MEMORY, verbose=0, fauxe2=None):
    '''
    Note: Only support s1 symmetry.
    '''
    assert comp == 1
    log = logger.new_logger(mol, verbose)
    t0 = (log._t0, log._w0)
    if auxmol is None:
        auxmol = addons.make_auxmol(mol, auxbasis)

    j2c = auxmol.intor(int2c, hermi=1)
    try:
        low = scipy.linalg.cholesky(j2c, lower=True)
        tag = 'cd'
    # pylint: disable=bare-except
    except: #scipy.linalg.LinAlgError:
        w, v = scipy.linalg.eigh(j2c)
        idx = w > LINEAR_DEP_THR
        low = (v[:,idx] / np.sqrt(w[idx]))
        v = None
        tag = 'eig'
    j2c = None
    naoaux, naux = low.shape
    log.debug('size of aux basis %d', naux)
    log.timer_debug1('2c2e', *t0)

    nao = mol.nao
    if aosym != 's1':
        raise NotImplementedError

    shls_slice = (0, mol.nbas, 0, mol.nbas, mol.nbas, mol.nbas+auxmol.nbas)
    ints = int3c_cross(mol, auxmol, intor=int3c, comp=comp, aosym=aosym,
                       out=None, shls_slice=shls_slice)
    ints = ints.reshape((-1,naoaux)).T

    if tag == 'cd':
        cderi = scipy.linalg.solve_triangular(low, ints, lower=True,
                                              overwrite_b=True, check_finite=False)
    else:
        cderi = np.dot(low.T, ints)

    log.timer('cholesky_eri', *t0)
    del log
    return cderi.reshape(-1,nao,nao)



from functools import reduce, partial
from jaxopt import linear_solve
from pyscf import __config__
from pyscf import numpy as np
from pyscf.lib import logger
from pyscf.cc import ccsd as pyscf_ccsd
from pyscf.mp.mp2 import _mo_without_core
from pyscfad import lib
from pyscfad.lib import jit
from pyscfad import util
from pyscfad.gto import mole
from pyscfad.scf import hf
from pyscfad import implicit_diff

CCSD_IMPLICIT_DIFF = getattr(__config__, "pyscfad_ccsd_implicit_diff", False)
# assume 'mol', 'mo_coeff', etc. come from '_scf',
# otherwise they need to be traced
CC_Tracers = ['_scf']
# attributes explicitly appearing in :fun:`update_amps` need to be traced
ERI_Tracers = ['fock', 'mo_energy', #'mol', 'mo_coeff', 'e_hf',
               'oooo', 'ooov', 'ovoo', 'ovov', 'oovv', 'ovvo', 'ovvv', 'vvvv']

def _converged_iter(amp, mycc, eris):
    t1, t2 = mycc.vector_to_amplitudes(amp)
    t1, t2 = mycc.update_amps(t1, t2, eris)
    amp = mycc.amplitudes_to_vector(t1, t2)
    return amp

def _iter(amp, mycc, eris, *,
          diis=None, max_cycle=50, tol=1e-8,
          tolnormt=1e-6, verbose=None):
    log = logger.new_logger(mycc, verbose)

    t1, t2 = mycc.vector_to_amplitudes(amp)
    eold = 0
    eccsd = mycc.energy(t1, t2, eris)
    log.info('Init E_corr(CCSD) = %.15g', eccsd)
    cput1 = log.timer('initialize CCSD')

    conv = False
    for istep in range(max_cycle):
        t1new, t2new = mycc.update_amps(t1, t2, eris)
        tmpvec = mycc.amplitudes_to_vector(t1new, t2new)
        tmpvec -= mycc.amplitudes_to_vector(t1, t2)
        normt = np.linalg.norm(tmpvec)
        tmpvec = None
        if mycc.iterative_damping < 1.0:
            alpha = mycc.iterative_damping
            t1new = (1-alpha) * t1 + alpha * t1new
            t2new *= alpha
            t2new += (1-alpha) * t2
        t1, t2 = t1new, t2new
        t1new = t2new = None
        t1, t2 = mycc.run_diis(t1, t2, istep, normt, eccsd-eold, diis)
        eold, eccsd = eccsd, mycc.energy(t1, t2, eris)
        log.info('cycle = %d  E_corr(CCSD) = %.15g  dE = %.9g  norm(t1,t2) = %.6g',
                 istep+1, eccsd, eccsd - eold, normt)
        cput1 = log.timer('CCSD iter', *cput1)
        if abs(eccsd-eold) < tol and normt < tolnormt:
            conv = True
            break
    amp = mycc.amplitudes_to_vector(t1, t2)
    return amp, conv

if CCSD_IMPLICIT_DIFF:
    solver = partial(linear_solve.solve_gmres, tol=1e-5,
                     solve_method='incremental', maxiter=30)
    _iter = implicit_diff.custom_fixed_point(
                _converged_iter, solve=solver, has_aux=True)(_iter)

def kernel(mycc, eris=None, t1=None, t2=None, max_cycle=50, tol=1e-8,
           tolnormt=1e-6, verbose=None):
    log = logger.new_logger(mycc, verbose)
    if eris is None:
        eris = mycc.ao2mo(mycc.mo_coeff)
    if t1 is None and t2 is None:
        t1, t2 = mycc.get_init_guess(eris)
    elif t2 is None:
        t2 = mycc.get_init_guess(eris)[1]

    cput1 = cput0 = (logger.process_clock(), logger.perf_counter())
    if isinstance(mycc.diis, lib.diis.DIIS):
        adiis = mycc.diis
    elif mycc.diis:
        adiis = lib.diis.DIIS(mycc, mycc.diis_file, incore=mycc.incore_complete)
        adiis.space = mycc.diis_space
    else:
        adiis = None

    vec = mycc.amplitudes_to_vector(t1, t2)
    vec, conv = _iter(vec, mycc, eris,
                      diis=adiis, max_cycle=max_cycle, tol=tol,
                      tolnormt=tolnormt, verbose=log)
    t1, t2 = mycc.vector_to_amplitudes(vec)
    eccsd = mycc.energy(t1, t2, eris)
    log.timer('CCSD', *cput0)
    return conv, eccsd, t1, t2

@jit
def amplitudes_to_vector(t1, t2, out=None):
    nocc, nvir = t1.shape
    nov = nocc * nvir
    vector_t1 = t1.ravel()
    vector_t2 = t2.transpose(0,2,1,3).reshape(nov,nov)[np.tril_indices(nov)]
    vector = np.concatenate((vector_t1, vector_t2), axis=None)
    return vector

def vector_to_amplitudes(vector, nmo, nocc):
    nvir = nmo - nocc
    nov = nocc * nvir
    t1 = vector[:nov].reshape((nocc,nvir))
    # filltriu=lib.SYMMETRIC because t2[iajb] == t2[jbia]
    t2 = lib.unpack_tril(vector[nov:], filltriu=lib.SYMMETRIC)
    t2 = t2.reshape(nocc,nvir,nocc,nvir).transpose(0,2,1,3)
    return t1, t2

@util.pytree_node(CC_Tracers, num_args=1)
class CCSD(pyscf_ccsd.CCSD):
    def __init__(self, mf, frozen=None, mo_coeff=None, mo_occ=None, **kwargs):
        pyscf_ccsd.CCSD.__init__(self, mf, frozen=frozen,
                                 mo_coeff=mo_coeff, mo_occ=mo_occ)
        if self.diis is True:
            self.diis = lib.diis.DIIS(self, self.diis_file, incore=self.incore_complete)
        self.__dict__.update(kwargs)

    def amplitudes_to_vector(self, t1, t2, out=None):
        return amplitudes_to_vector(t1, t2, out)

    def vector_to_amplitudes(self, vec, nmo=None, nocc=None):
        if nocc is None: nocc = self.nocc
        if nmo is None: nmo = self.nmo
        return vector_to_amplitudes(vec, nmo, nocc)

    def ccsd(self, t1=None, t2=None, eris=None):
        assert(self.mo_coeff is not None)
        assert(self.mo_occ is not None)

        if self.verbose >= logger.WARN:
            self.check_sanity()
        self.dump_flags()

        if eris is None:
            eris = self.ao2mo(self.mo_coeff)

        self.e_hf = getattr(eris, 'e_hf', None)
        if self.e_hf is None:
            self.e_hf = self._scf.e_tot

        self.converged, self.e_corr, self.t1, self.t2 = \
                kernel(self, eris, t1, t2, max_cycle=self.max_cycle,
                       tol=self.conv_tol, tolnormt=self.conv_tol_normt,
                       verbose=self.verbose)
        self._finalize()
        return self.e_corr, self.t1, self.t2

    def ccsd_t(self, t1=None, t2=None, eris=None):
        from pyscfad.cc import ccsd_t_slow
        if t1 is None: t1 = self.t1
        if t2 is None: t2 = self.t2
        if eris is None: eris = self.ao2mo(self.mo_coeff)
        return ccsd_t_slow.kernel(self, eris, t1, t2, self.verbose)

    def ipccsd(self, nroots=1, left=False, koopmans=False, guess=None,
               partition=None, eris=None):
        from pyscfad.cc import eom_rccsd
        return eom_rccsd.EOMIP(self).kernel(nroots, left, koopmans, guess,
                                            partition, eris)

    def eaccsd(self, nroots=1, left=False, koopmans=False, guess=None,
               partition=None, eris=None):
        from pyscfad.cc import eom_rccsd
        return eom_rccsd.EOMEA(self).kernel(nroots, left, koopmans, guess,
                                            partition, eris)

    def eeccsd(self, nroots=1, koopmans=False, guess=None, eris=None):
        from pyscfad.cc import eom_rccsd
        return eom_rccsd.EOMEE(self).kernel(nroots, koopmans, guess, eris)

@util.pytree_node(ERI_Tracers)
class _ChemistsERIs(pyscf_ccsd._ChemistsERIs):
    def __init__(self, mol=None, **kwargs):
        pyscf_ccsd._ChemistsERIs.__init__(self, mol=mol)
        self.__dict__.update(kwargs)

    def _common_init_(self, mycc, mo_coeff=None):
        if mo_coeff is None:
            mo_coeff = mycc.mo_coeff
        self.mo_coeff = mo_coeff = _mo_without_core(mycc, mo_coeff)

        dm = mycc._scf.make_rdm1(mycc.mo_coeff, mycc.mo_occ)
        vhf = mycc._scf.get_veff(mycc.mol, dm)
        fockao = mycc._scf.get_fock(vhf=vhf, dm=dm)
        self.fock = reduce(np.dot, (mo_coeff.conj().T, fockao, mo_coeff))
        self.e_hf = mycc._scf.energy_tot(dm=dm, vhf=vhf)
        nocc = self.nocc = mycc.nocc
        self.mol = mycc.mol

        mo_e = self.mo_energy = self.fock.diagonal().real
        return self



from functools import partial
import numpy
from jax import jacrev, jacfwd
from jaxopt import linear_solve
from pyscf import __config__
from pyscf import numpy as np
from pyscf.lib import param, logger
from pyscf.scf import hf as pyscf_hf
from pyscf.scf import chkfile
from pyscf.scf.hf import TIGHT_GRAD_CONV_TOL
from pyscfad import lib
from pyscfad.lib import jit
from pyscfad import util
from pyscfad import implicit_diff
from pyscfad import gto
from pyscfad import df
from pyscfad.lib import stop_grad
from pyscfad.scf import _vhf
from pyscfad.scf.diis import SCF_DIIS
from pyscfad.scf._eigh import eigh

SCF_IMPLICIT_DIFF = getattr(__config__, "pyscfad_scf_implicit_diff", False)
Traced_Attributes = ['mol', 'mo_coeff', 'mo_energy', '_eri']

def eig(h, s, x0=None):
    e, c = eigh(h, s, x0)
    return e, c

def _converged_scf(mo_coeff, mf, s1e, h1e, mo_occ):
    mol = getattr(mf, "cell", mf.mol)
    dm = mf.make_rdm1(mo_coeff, mo_occ)
    vhf = mf.get_veff(mol, dm)
    fock = mf.get_fock(h1e, s1e, vhf, dm)
    _, mo_coeff = mf.eig(fock, s1e, mo_coeff)
    return mo_coeff

def _scf(mo_coeff, mf, s1e, h1e, mo_occ, *,
         dm0=None, conv_tol=1e-10, conv_tol_grad=None, diis=None,
         dump_chk=True, callback=None, log=None):
    if conv_tol_grad is None:
        conv_tol_grad = numpy.sqrt(conv_tol)
    if log is None:
        log = logger.new_logger(mf)
    scf_conv = False

    mol = getattr(mf, "cell", mf.mol)
    if dm0 is not None:
        dm = dm0
    elif mo_coeff is not None and mo_occ is not None:
        dm = mf.make_rdm1(mo_coeff, mo_occ)
    else:
        raise KeyError("Either dm or mo_coeff and mo_occ need to be given.")

    vhf = mf.get_veff(mol, dm)
    e_tot = mf.energy_tot(dm, h1e, vhf)
    log.info('init E= %.15g', e_tot)

    cput1 = log.timer('initialize scf')
    for cycle in range(mf.max_cycle):
        dm_last = dm
        last_hf_e = e_tot

        fock = mf.get_fock(h1e, s1e, vhf, dm, cycle, diis)
        mo_energy, mo_coeff = mf.eig(fock, s1e)
        mo_occ = mf.get_occ(stop_grad(mo_energy), stop_grad(mo_coeff))
        dm = mf.make_rdm1(mo_coeff, mo_occ)
        vhf = mf.get_veff(mol, dm, dm_last, vhf)
        e_tot = mf.energy_tot(dm, h1e, vhf)

        fock = mf.get_fock(stop_grad(h1e), stop_grad(s1e), stop_grad(vhf), stop_grad(dm))
        norm_gorb = numpy.linalg.norm(mf.get_grad(stop_grad(mo_coeff),
                                      stop_grad(mo_occ), stop_grad(fock)))
        if not TIGHT_GRAD_CONV_TOL:
            norm_gorb = norm_gorb / numpy.sqrt(norm_gorb.size)
        norm_ddm = numpy.linalg.norm(stop_grad(dm)-stop_grad(dm_last))
        log.info('cycle= %d E= %.15g  delta_E= %4.3g  |g|= %4.3g  |ddm|= %4.3g',
                 cycle+1, e_tot, e_tot-last_hf_e, norm_gorb, norm_ddm)

        if callable(mf.check_convergence):
            scf_conv = mf.check_convergence(locals())
        elif abs(e_tot-last_hf_e) < conv_tol and norm_gorb < conv_tol_grad:
            scf_conv = True

        if dump_chk:
            mf.dump_chk(locals())

        if callable(callback):
            callback(locals())

        cput1 = log.timer('cycle= %d'%(cycle+1), *cput1)

        if scf_conv:
            break
    return mo_coeff, scf_conv, mo_occ, mo_energy

if SCF_IMPLICIT_DIFF:
    solver = partial(linear_solve.solve_gmres, tol=1e-9,
                     solve_method='incremental', maxiter=30)
    _scf = implicit_diff.custom_fixed_point(
                _converged_scf, solve=solver, has_aux=True,
                nondiff_argnums=(4,), use_converged_args={4:2})(_scf)


def kernel(mf, conv_tol=1e-10, conv_tol_grad=None,
           dump_chk=True, dm0=None, callback=None, conv_check=True, **kwargs):
    log = logger.new_logger(mf)
    cput0 = (log._t0, log._w0)
    if conv_tol_grad is None:
        conv_tol_grad = numpy.sqrt(conv_tol)
        log.info('Set gradient conv threshold to %g', conv_tol_grad)

    mol = getattr(mf, "cell", mf.mol)
    if dm0 is None:
        dm = mf.get_init_guess(mol, mf.init_guess)
    else:
        dm = dm0

    h1e = mf.get_hcore(mol)
    if mf._eri is None:
        if getattr(mf, 'with_df', None) is None:
            mf._eri = mol.intor('int2e', aosym='s1')

    scf_conv = False
    mo_energy = mo_coeff = mo_occ = None

    s1e = mf.get_ovlp(mol)
    cond = numpy.linalg.cond(stop_grad(s1e))
    log.debug('cond(S) = %s', cond)
    if cond.max()*1e-17 > conv_tol:
        log.warn('Singularity detected in overlap matrix (condition number = %4.3g). '
                 'SCF may be inaccurate and hard to converge.', cond.max())

    if mf.max_cycle <= 0:
        # Skip SCF iterations. Compute only the total energy of the initial density
        vhf = mf.get_veff(mol, dm)
        e_tot = mf.energy_tot(dm, h1e, vhf)
        log.info('init E= %.15g', e_tot)

        fock = mf.get_fock(h1e, s1e, vhf, dm)
        mo_energy, mo_coeff = mf.eig(fock, s1e)
        mo_occ = mf.get_occ(stop_grad(mo_energy), stop_grad(mo_coeff))
        return scf_conv, e_tot, mo_energy, mo_coeff, mo_occ

    if isinstance(mf.diis, lib.diis.DIIS):
        mf_diis = mf.diis
    elif mf.diis:
        assert issubclass(mf.DIIS, lib.diis.DIIS)
        mf_diis = mf.DIIS(mf, mf.diis_file)
        mf_diis.space = mf.diis_space
        mf_diis.rollback = mf.diis_space_rollback
    else:
        mf_diis = None

    if dump_chk and mf.chkfile:
        # Explicit overwrite the mol object in chkfile
        # Note in pbc.scf, mf.mol == mf.cell, cell is saved under key "mol"
        chkfile.save_mol(mol, mf.chkfile)

    # A preprocessing hook before the SCF iteration
    #mf.pre_kernel(locals())

    # SCF iteration
    mo_coeff, scf_conv, mo_occ, mo_energy = \
            _scf(mo_coeff, mf, s1e, h1e, mo_occ, dm0=dm,
                 conv_tol=conv_tol, conv_tol_grad=conv_tol_grad,
                 diis=mf_diis, dump_chk=dump_chk, callback=callback, log=log)

    # Recompute energy so that energy has the correct gradient from mo_coeff
    dm = mf.make_rdm1(mo_coeff, mo_occ)
    vhf = mf.get_veff(mol, dm)
    e_tot = mf.energy_tot(dm, h1e, vhf)

    run_extra_cycle = False
    if SCF_IMPLICIT_DIFF and (not conv_check or not scf_conv):
        log.warn('\tAn extra scf cycle is going to be run\n'
                 '\tin order to restore the mo_energy derivatives\n'
                 '\tmissing in implicit differentiation.')
        run_extra_cycle = True

    if (scf_conv and conv_check) or run_extra_cycle:
        # An extra diagonalization, to remove level shift
        fock = mf.get_fock(h1e, s1e, vhf, dm)
        mo_energy, mo_coeff = mf.eig(fock, s1e)
        mo_occ = mf.get_occ(stop_grad(mo_energy), stop_grad(mo_coeff))
        dm, dm_last = mf.make_rdm1(mo_coeff, mo_occ), dm
        vhf = mf.get_veff(mol, dm, dm_last, vhf)
        e_tot, last_hf_e = mf.energy_tot(dm, h1e, vhf), e_tot

        fock = mf.get_fock(stop_grad(h1e), stop_grad(s1e),
                           stop_grad(vhf), stop_grad(dm))
        norm_gorb = numpy.linalg.norm(mf.get_grad(stop_grad(mo_coeff),
                                      stop_grad(mo_occ), stop_grad(fock)))
        if not TIGHT_GRAD_CONV_TOL:
            norm_gorb = norm_gorb / numpy.sqrt(norm_gorb.size)
        norm_ddm = numpy.linalg.norm(stop_grad(dm)-stop_grad(dm_last))

        conv_tol = conv_tol * 10
        conv_tol_grad = conv_tol_grad * 3
        if callable(mf.check_convergence):
            scf_conv = mf.check_convergence(locals())
        elif abs(e_tot-last_hf_e) < conv_tol or norm_gorb < conv_tol_grad:
            scf_conv = True
        log.info('Extra cycle  E= %.15g  delta_E= %4.3g  |g|= %4.3g  |ddm|= %4.3g',
                 e_tot, e_tot-last_hf_e, norm_gorb, norm_ddm)
        if dump_chk:
            mf.dump_chk(locals())

    log.timer('scf_cycle', *cput0)
    del log
    # A post-processing hook before return
    #mf.post_kernel(locals())
    return scf_conv, e_tot, mo_energy, mo_coeff, mo_occ

def dot_eri_dm(eri, dm, hermi=0, with_j=True, with_k=True):
    dm = np.asarray(dm)
    nao = dm.shape[-1]
    if eri.dtype == np.complex128 or eri.size == nao**4:
        vj, vk = _dot_eri_dm_nosymm(eri, dm, with_j, with_k)
    else:
        if dm.dtype == np.complex128:
            raise NotImplementedError
        vj, vk = _vhf.incore(eri, dm, hermi, with_j, with_k)
    return vj, vk

@partial(jit, static_argnums=(2,3))
def _dot_eri_dm_nosymm(eri, dm, with_j, with_k):
    nao = dm.shape[-1]
    eri = eri.reshape((nao,)*4)
    dms = dm.reshape(-1,nao,nao)
    vj = vk = None
    if with_j:
        vj = np.einsum('ijkl,xji->xkl', eri, dms)
        vj = vj.reshape(dm.shape)
    if with_k:
        vk = np.einsum('ijkl,xjk->xil', eri, dms)
        vk = vk.reshape(dm.shape)
    return vj, vk

@util.pytree_node(Traced_Attributes, num_args=1)
class SCF(pyscf_hf.SCF):
    '''
    A subclass of :class:`pyscf.scf.hf.SCF` where the following
    attributes can be traced.

    Attributes:
        mol : :class:`pyscfad.gto.Mole` object
            Molecular structure and global options.
        mo_coeff : array
            Molecular orbital coefficients.
        mo_energy : array
            Molecular orbital energies.
        _eri : array
            Two electron repulsion integrals.
    '''
    DIIS = SCF_DIIS

    def __init__(self, mol, **kwargs):
        pyscf_hf.SCF.__init__(self, mol)
        self.__dict__.update(kwargs)

    def get_jk(self, mol=None, dm=None, hermi=1, with_j=True, with_k=True,
               omega=None):
        if mol is None:
            mol = self.mol
        if dm is None:
            dm = self.make_rdm1()
        if self._eri is None:
            self._eri = self.mol.intor('int2e', aosym='s1')
        vj, vk = dot_eri_dm(self._eri, dm, hermi, with_j, with_k)
        return vj, vk

    def get_init_guess(self, mol=None, key='minao'):
        if mol is None:
            mol = self.mol
        mol = stop_grad(mol)
        return pyscf_hf.SCF.get_init_guess(self, mol, key)

    # pylint: disable=arguments-differ
    def kernel(self, dm0=None, **kwargs):
        self.build(self.mol)
        self.converged, self.e_tot, \
                self.mo_energy, self.mo_coeff, self.mo_occ = \
                kernel(self, self.conv_tol, self.conv_tol_grad,
                       dm0=dm0, **kwargs)
        return self.e_tot

    def _eigh(self, h, s, x0=None):
        return eig(h, s, x0)

    def eig(self, h, s, x0=None):
        return self._eigh(h, s, x0)

    def energy_grad(self, dm0=None, mode="rev"):
        """
        Energy gradient wrt AO parameters computed by AD.
        In principle, MO response is not needed, and we can just take
        the derivative of eigen decomposition with converged
        density matrix. But here it is implemented in this way to show
        the difference between unrolling for loops and implicit differentiation.

        NOTE:
            The attributes of the SCF instance will not be modified
        """
        if dm0 is None:
            try:
                dm0 = self.make_rdm1()
            except TypeError:
                pass

        def hf_energy(self, dm0=None):
            self.reset()
            e_tot = self.kernel(dm0=dm0)
            return e_tot

        if mode == "rev":
            jac = jacrev(hf_energy)(self, dm0=dm0)
        else:
            if SCF_IMPLICIT_DIFF:
                msg = """Forward mode differentiation is not available
                         when applying the implicit function differentiation."""
                raise KeyError(msg)
            jac = jacfwd(hf_energy)(self, dm0=dm0)
        if hasattr(jac,"cell"):
            return jac.cell
        else:
            return jac.mol

    def density_fit(self, auxbasis=None, with_df=None, only_dfj=False):
        return df.density_fit(self, auxbasis, with_df, only_dfj)

@util.pytree_node(Traced_Attributes, num_args=1)
class RHF(SCF, pyscf_hf.RHF):
    pass



import warnings
from functools import partial
import ctypes
import numpy
from jax import jit
from jax import vmap
from pyscf import ao2mo
from pyscf.gto import mole, moleintor
from pyscf.gto.mole import Mole
from pyscf.gto.moleintor import _get_intor_and_comp
from pyscfad.lib import numpy as np
from pyscfad.lib import ops, custom_jvp
from ._mole_helper import uncontract, setup_exp, setup_ctr_coeff

SET_RC = ["rinv",]

@partial(custom_jvp, nondiff_argnums=(0,3,4))
def intor_cross(intor, mol1, mol2, comp=None, grids=None):
    return mole.intor_cross(intor, mol1, mol2, comp=comp, grids=grids)

@intor_cross.defjvp
def intor_cross_jvp(intor, comp, grids,
                    primals, tangents):
    mol1, mol2 = primals
    mol1_t, mol2_t = tangents

    primal_out = intor_cross(intor, mol1, mol2, comp=comp, grids=grids)
    tangent_out = np.zeros_like(primal_out)

    if mol1.coords is not None:
        nao1 = mol1.nao
        aoslices1 = mol1.aoslice_by_atom()[:,2:4]
        intor_ip_bra, _ = _int1e_dr1_name(intor)
        s1a = -intor_cross(intor_ip_bra, mol1, mol2, comp=None, grids=grids).reshape(3,-1,nao1,nao1)

        idx1 = np.arange(nao1)
        grad1 = _gen_int1e_fill_grad_r0(s1a, aoslices1, idx1[None,None,:,None])
        tangent_out += np.einsum('nxyij,nx->yij', grad1, mol1_t.coords).reshape(primal_out.shape)

    if mol2.coords is not None:
        nao2 = mol2.nao
        aoslices2 = mol2.aoslice_by_atom()[:,2:4]
        _, intor_ip_ket = _int1e_dr1_name(intor)
        s1b = -intor_cross(intor_ip_ket, mol1, mol2, comp=None, grids=grids)

        order_a = _int1e_get_dr_order(intor_ip_ket)[0]
        s1b = s1b.reshape(3**order_a,3,-1,nao2,nao2).transpose(1,0,2,3,4).reshape(3,-1,nao2,nao2)
        idx2 = np.arange(nao2)
        grad2 = _gen_int1e_fill_grad_r0(s1b, aoslices2, idx2[None,None,None,:])
        tangent_out += np.einsum('nxyij,nx->yij', grad2, mol2_t.coords).reshape(primal_out.shape)

    if mol1.ctr_coeff is not None:
        pass
    if mol1.exp is not None:
        pass

    if mol2.ctr_coeff is not None:
        pass
    if mol2.exp is not None:
        pass

    return primal_out, tangent_out

def getints(mol, intor, shls_slice=None,
            comp=None, hermi=0, aosym='s1', out=None):
    if intor.endswith("_spinor"):
        raise NotImplementedError('Spinors are not supported for AD.')
    if hermi == 2:
        hermi = 0
        msg = f'Anti-hermitian symmetry is not supported. Setting hermi = {hermi}.'
        warnings.warn(msg)
    if aosym != 's1':
        aosym = 's1'
        msg = f'AO symmetry is not supported. Setting aosym = {aosym}.'
        warnings.warn(msg)

    if (intor.startswith("int1e") or
        intor.startswith("int2c2e") or
        intor.startswith('ECP')):
        return getints2c(mol, intor, shls_slice, comp, hermi, aosym, out)
    elif intor.startswith("int2e"):
        return getints4c(mol, intor, shls_slice, comp, aosym, out)
    else:
        raise NotImplementedError

def _int1e_dr1_name(intor):
    if 'sph' in intor:
        suffix = '_sph'
    elif 'cart' in intor:
        suffix = '_cart'
    else:
        suffix = ''
    fname = intor.replace('_sph', '').replace('_cart', '')

    if fname[-4:-2] == 'dr':
        orders = [int(fname[-2]), int(fname[-1])]
        intor_ip_bra = fname[:-2] + str(orders[0]+1) + str(orders[1]) + suffix
        intor_ip_ket = fname[:-2] + str(orders[0]) + str(orders[1]+1) + suffix
    else:
        intor_ip_bra = fname + '_dr10' + suffix
        intor_ip_ket = fname + '_dr01' + suffix
    return intor_ip_bra, intor_ip_ket

def _int1e_get_dr_order(intor):
    fname = intor.replace('_sph', '').replace('_cart', '')
    if fname[-4:-2] == 'dr':
        orders = [int(fname[-2]), int(fname[-1])]
    else:
        orders = [0, 0]
    return orders

def _int2e_dr1_name(intor):
    if 'sph' in intor:
        suffix = '_sph'
    elif 'cart' in intor:
        suffix = '_cart'
    else:
        suffix = ''
    fname = intor.replace('_sph', '').replace('_cart', '')

    if fname[-6:-4] == 'dr':
        orders = _int2e_get_dr_order(intor)
        str1 = str(orders[0]+1) + str(orders[1]) + str(orders[2]) + str(orders[3])
        str2 = str(orders[0]) + str(orders[1]+1) + str(orders[2]) + str(orders[3])
        str3 = str(orders[0]) + str(orders[1]) + str(orders[2]+1) + str(orders[3])
        str4 = str(orders[0]) + str(orders[1]) + str(orders[2]) + str(orders[3]+1)
        intor1 = fname[:-4] + str1 + suffix
        intor2 = fname[:-4] + str2 + suffix
        intor3 = fname[:-4] + str3 + suffix
        intor4 = fname[:-4] + str4 + suffix
    else:
        intor1 = fname + '_dr1000' + suffix
        intor2 = fname + '_dr0100' + suffix
        intor3 = fname + '_dr0010' + suffix
        intor4 = fname + '_dr0001' + suffix
    return intor1, intor2, intor3, intor4

def _int2e_get_dr_order(intor):
    fname = intor.replace('_sph', '').replace('_cart', '')
    if fname[-6:-4] == 'dr':
        orders = [int(fname[-4]), int(fname[-3]), int(fname[-2]), int(fname[-1])]
    else:
        orders = [0,] * 4
    return orders

def getints2c_rc(mol, intor, shls_slice=None, comp=None,
                 hermi=0, aosym='s1', out=None, rc_deriv=None):
    if rc_deriv is None or not any(rc in intor for rc in SET_RC):
        return getints2c(mol, intor, shls_slice, comp, hermi, aosym, out)
    else:
        return _getints2c_rc(mol, intor, shls_slice, comp, hermi, aosym, out, rc_deriv)

@partial(custom_jvp, nondiff_argnums=tuple(range(1,8)))
def _getints2c_rc(mol, intor, shls_slice=None, comp=None,
                  hermi=0, aosym='s1', out=None, rc_deriv=None):
    return Mole.intor(mol, intor, comp=comp, hermi=hermi, aosym=aosym,
                      shls_slice=shls_slice, out=out)

@_getints2c_rc.defjvp
def _getints2c_rc_jvp(intor, shls_slice, comp, hermi, aosym, out, rc_deriv,
                      primals, tangents):
    if shls_slice is not None:
        raise NotImplementedError

    mol, = primals
    mol_t, = tangents
    primal_out = _getints2c_rc(mol, intor, shls_slice, comp,
                               hermi, aosym, out)
    tangent_out = np.zeros_like(primal_out)

    if mol.coords is not None:
        intor_ip_bra, intor_ip_ket = _int1e_dr1_name(intor)
        tangent_out += _gen_int1e_jvp_r0(mol, mol_t, intor_ip_bra, intor_ip_ket, rc_deriv)
    return primal_out, tangent_out

@partial(custom_jvp, nondiff_argnums=tuple(range(1,7)))
def getints2c(mol, intor, shls_slice=None, comp=None, hermi=0, aosym='s1', out=None):
    return Mole.intor(mol, intor, comp=comp, hermi=hermi, aosym=aosym,
                      shls_slice=shls_slice, out=out)

@getints2c.defjvp
def getints2c_jvp(intor, shls_slice, comp, hermi, aosym, out,
                  primals, tangents):
    if shls_slice is not None:
        msg = 'AD for integrals with subblock of shells are not supported yet.'
        raise NotImplementedError(msg)

    mol, = primals
    mol_t, = tangents

    primal_out = getints2c(mol, intor, shls_slice=shls_slice,
                           comp=comp, hermi=hermi, aosym=aosym, out=out)

    tangent_out = np.zeros_like(primal_out)
    fname = intor.replace('_sph', '').replace('_cart', '')
    if mol.coords is not None:
        intor_ip_bra = intor_ip_ket = intor_ip = None
        if intor.startswith("ECPscalar"):
            intor_ip = intor.replace("ECPscalar", "ECPscalar_ipnuc")
        elif fname == 'int1e_r':
            intor_ip = intor.replace('int1e_r', 'int1e_irp')
        elif fname.startswith("int1e") or fname.startswith("int2c2e"):
            intor_ip_bra, intor_ip_ket = _int1e_dr1_name(intor)
        else:
            raise NotImplementedError(f'Integral {intor} is not supported for AD.')

        if intor_ip_bra or intor_ip_ket:
            tangent_out += _gen_int1e_jvp_r0(mol, mol_t,
                                intor_ip_bra, intor_ip_ket, hermi=hermi).reshape(tangent_out.shape)
            if "nuc" in intor_ip_bra and "nuc" in intor_ip_ket:
                intor_ip_bra = intor_ip_bra.replace("nuc", "rinv")
                intor_ip_ket = intor_ip_ket.replace("nuc", "rinv")
                tangent_out += _gen_int1e_nuc_jvp_rc(mol, mol_t,
                                intor_ip_bra, intor_ip_ket, hermi=hermi).reshape(tangent_out.shape)
        elif fname == 'int1e_r':
            tangent_out += _int1e_r_jvp_r0(mol, mol_t, intor_ip)
        else:
            tangent_out += _int1e_jvp_r0(mol, mol_t, intor_ip)

        intor_ip = None
        if intor.startswith("ECPscalar"):
            intor_ip = intor.replace("ECPscalar", "ECPscalar_iprinv")
        if intor_ip:
            tangent_out += _int1e_nuc_jvp_rc(mol, mol_t, intor_ip)

    if mol.ctr_coeff is not None:
        tangent_out += _int1e_jvp_cs(mol, mol_t, intor)

    if mol.exp is not None:
        tangent_out += _int1e_jvp_exp(mol, mol_t, intor)
    return primal_out, tangent_out

@partial(custom_jvp, nondiff_argnums=tuple(range(1,6)))
def getints4c(mol, intor,
              shls_slice=None, comp=None, aosym='s1', out=None):
    if (shls_slice is None and aosym=='s1'
            and intor in ['int2e', 'int2e_sph', 'int2e_cart']):
        eri8 = Mole.intor(mol, intor, comp=comp, aosym='s8',
                          shls_slice=shls_slice, out=out)
        eri = ao2mo.restore(aosym, eri8, mol.nao)
        del eri8
    else:
        eri = Mole.intor(mol, intor, comp=comp, aosym=aosym,
                         shls_slice=shls_slice, out=out)
    return eri

@getints4c.defjvp
def getints4c_jvp(intor, shls_slice, comp, aosym, out,
                  primals, tangents):
    if shls_slice is not None or out is not None:
        raise NotImplementedError
    if aosym != 's1':
        raise NotImplementedError

    mol, = primals
    primal_out = getints4c(mol, intor, shls_slice, comp, aosym, out)

    mol_t, = tangents
    tangent_out = np.zeros_like(primal_out)

    fname = intor.replace('_sph', '').replace('_cart', '')
    if mol.coords is not None:
        intor1, intor2, intor3, intor4 = _int2e_dr1_name(intor)
        if fname[-6:-4] == 'dr':
            orders = _int2e_get_dr_order(intor)
            if orders[0] == 0 and orders[1] == 0:
                intor2 = None
            if orders[2] == 0 and orders[3] == 0:
                intor4 = None
            tangent_out += _gen_int2e_jvp_r0(mol, mol_t, [intor1, intor2, intor3, intor4])
        else:
            tangent_out += _int2e_jvp_r0(mol, mol_t, intor1)

    if mol.ctr_coeff is not None:
        tangent_out += _int2e_jvp_cs(mol, mol_t, intor)
    if mol.exp is not None:
        tangent_out += _int2e_jvp_exp(mol, mol_t, intor)
    return primal_out, tangent_out

def _get_fakemol_cs(mol):
    mol1 = uncontract(mol)
    return mol1

def _get_fakemol_exp(mol):
    mol1 = uncontract(mol)
    mol1._bas[:,mole.ANG_OF]  += 2
    return mol1

def get_bas_label(l):
    xyz = []
    for x in range(l, -1, -1):
        for y in range(l-x, -1, -1):
            z = l-x-y
            xyz.append('x'*x + 'y'*y + 'z'*z)
    return xyz

def promote_xyz(xyz, x, l):
    if x == 'z':
        return xyz+'z'*l
    elif x == 'y':
        if 'z' in xyz:
            tmp = xyz.split('z', 1)
            return tmp[0] + 'y'*l + 'z' + tmp[1]
        else:
            return xyz+'y'*l
    elif x == 'x':
        return 'x'*l+xyz
    else:
        raise ValueError

def _int1e_jvp_r0(mol, mol_t, intor):
    s1 = -getints2c(mol, intor, comp=3)
    aoslices = mol.aoslice_by_atom()[:,2:4]
    grad = _fill_grad_r0(s1, aoslices)
    tangent_out = _int1e_dot_grad_tangent_r0(grad, mol_t.coords)
    return tangent_out

@jit
def _fill_grad_r0(eri1, aoslices):
    nao = eri1.shape[-1]
    shape = [1,] * eri1.ndim
    shape[1] = nao
    idx = np.arange(nao)
    idx = idx.reshape(shape)
    def body(slices):
        p0, p1 = slices[:]
        mask = (idx >= p0) & (idx < p1)
        return np.where(mask, eri1, np.array(0, dtype=eri1.dtype))
    grad = vmap(body)(aoslices)
    return grad

@jit
def _int1e_dot_grad_tangent_r0(grad, tangent):
    tangent_out = np.einsum('nxij,nx->ij', grad, tangent)
    tangent_out += tangent_out.T
    return tangent_out

def _gen_int1e_jvp_r0(mol, mol_t, intor_a, intor_b, rc_deriv=None, hermi=0):
    coords_t = mol_t.coords
    #atmlst = range(mol.natm)
    aoslices = mol.aoslice_by_atom()[:,2:4]
    nao = mol.nao

    s1a = -getints2c_rc(mol, intor_a, hermi=0, rc_deriv=rc_deriv).reshape(3,-1,nao,nao)
    s1b = None
    if hermi == 0:
        order_a = _int1e_get_dr_order(intor_b)[0]
        s1b = -getints2c_rc(mol, intor_b, hermi=0, rc_deriv=rc_deriv)
        s1b = s1b.reshape(3**order_a,3,-1,nao,nao).transpose(1,0,2,3,4).reshape(3,-1,nao,nao)
    #jvp = np.zeros(s1a.shape[1:])
    #for k, ia in enumerate(atmlst):
    #    p0, p1 = aoslices[ia,2:]
    #    ta = np.einsum('xyij,x->yij', s1a[...,p0:p1,:], coords_t[k])
    #    tb = np.einsum('xyij,x->yij', s1b[...,p0:p1], coords_t[k])
    #    jvp = ops.index_add(jvp, ops.index[:,p0:p1], ta)
    #    jvp = ops.index_add(jvp, ops.index[:,:,p0:p1], tb)
    idx = np.arange(nao)
    grad = _gen_int1e_fill_grad_r0(s1a, aoslices, idx[None,None,:,None])
    if hermi == 0:
        grad = grad + _gen_int1e_fill_grad_r0(s1b, aoslices, idx[None,None,None,:])
    if rc_deriv is not None:
        grad_rc = -s1a
        if hermi == 0:
            grad_rc -= s1b
        grad = ops.index_add(grad, ops.index[rc_deriv], grad_rc)
    jvp = np.einsum('nxyij,nx->yij', grad, coords_t)
    if hermi == 1:
        jvp += jvp.transpose(0,2,1)
    return jvp

@jit
def _gen_int1e_fill_grad_r0(s1, aoslices, idx):
    def body(slices):
        p0, p1 = slices[:]
        mask = (idx >= p0) & (idx < p1)
        grad = np.where(mask, s1, np.array(0, dtype=s1.dtype))
        return grad
    grad = vmap(body)(aoslices)
    return grad

def _int1e_r_jvp_r0(mol, mol_t, intor):
    coords_t = mol_t.coords
    atmlst = range(mol.natm)
    aoslices = mol.aoslice_by_atom()
    nao = mol.nao
    s1 = -getints2c(mol, intor).reshape(-1,3,nao,nao)
    grad = [numpy.zeros_like(s1) for ia in atmlst]
    grad = numpy.asarray(grad)
    for k, ia in enumerate(atmlst):
        p0, p1 = aoslices [ia,2:]
        grad[k,...,p0:p1] = s1[...,p0:p1]
    tangent_out = _int1e_r_dot_grad_tangent_r0(grad, coords_t)
    return tangent_out

@jit
def _int1e_r_dot_grad_tangent_r0(grad, tangent):
    tangent_out = np.einsum('npxij,nx->pij', grad, tangent)
    tangent_out += tangent_out.transpose(0,2,1)
    return tangent_out

def _int1e_nuc_jvp_rc(mol, mol_t, intor):
    coords_t = mol_t.coords
    atmlst = range(mol.natm)
    nao = mol.nao
    grad = np.zeros((mol.natm,3,nao,nao), dtype=float)
    for k, ia in enumerate(atmlst):
        with mol.with_rinv_at_nucleus(ia):
            vrinv = getints2c_rc(mol, intor, comp=3, rc_deriv=ia)
            if "ECP" not in intor:
                vrinv *= -mol.atom_charge(ia)
        grad = ops.index_update(grad, ops.index[k], vrinv)
    tangent_out = _int1e_dot_grad_tangent_r0(grad, coords_t)
    return tangent_out

def _gen_int1e_nuc_jvp_rc(mol, mol_t, intor_a, intor_b, hermi=0):
    coords_t = mol_t.coords
    atmlst = range(mol.natm)
    nao = mol.nao
    _, comp = _get_intor_and_comp(intor_a)
    grad = np.zeros((mol.natm,3,comp//3,nao,nao), dtype=float)
    for k, ia in enumerate(atmlst):
        with mol.with_rinv_at_nucleus(ia):
            vrinv = getints2c_rc(mol, intor_a, rc_deriv=ia).reshape(3,-1,nao,nao)
            if hermi == 0:
                order_a = _int1e_get_dr_order(intor_b)[0]
                s1b = getints2c_rc(mol, intor_b, rc_deriv=ia)
                s1b = s1b.reshape(3**order_a,3,-1,nao,nao).transpose(1,0,2,3,4)
                vrinv += s1b.reshape(3,-1,nao,nao)
            if "ECP" not in intor_a:
                vrinv *= -mol.atom_charge(ia)
        grad = ops.index_update(grad, ops.index[k], vrinv)
    jvp = np.einsum('nxyij,nx->yij', grad, coords_t)
    if hermi == 1:
        jvp = jvp + jvp.transpose(0,2,1)
    return jvp

def _int1e_jvp_cs(mol, mol_t, intor):
    ctr_coeff = mol.ctr_coeff
    ctr_coeff_t = mol_t.ctr_coeff

    mol1 = _get_fakemol_cs(mol)
    mol1._atm[:,mole.CHARGE_OF] = 0 # set nuclear charge to zero
    nbas1 = len(mol1._bas)
    nbas = len(mol._bas)
    shls_slice = (0, nbas1, nbas1, nbas1+nbas)
    intor = mol._add_suffix(intor)
    if 'ECP' in intor:
        assert mol._ecp is not None
        bas = numpy.vstack((mol._bas, mol._ecpbas))
    else:
        bas = mol._bas
    atmc, basc, envc = mole.conc_env(mol1._atm, mol1._bas, mol1._env,
                                     mol._atm, bas, mol._env)
    if 'ECP' in intor:
        envc[mole.AS_ECPBAS_OFFSET] = len(mol1._bas) + len(mol._bas)
        envc[mole.AS_NECPBAS] = len(mol._ecpbas)

    s = moleintor.getints(intor, atmc, basc, envc, shls_slice)
    _, cs_of, _ = setup_ctr_coeff(mol)
    nao = mol.nao
    #grad = np.zeros((len(ctr_coeff), nao, nao), dtype=float)
    grad = numpy.zeros((len(ctr_coeff), nao, nao), dtype=float)

    off = 0
    ibas = 0
    for i in range(len(mol._bas)):
        l = mol._bas[i,mole.ANG_OF]
        if mol.cart:
            nbas = (l+1)*(l+2)//2
        else:
            nbas = 2*l + 1
        nprim = mol._bas[i,mole.NPRIM_OF]
        nctr = mol._bas[i,mole.NCTR_OF]
        g = s[off:(off+nprim*nbas)].reshape(nprim,-1,nao)
        for j in range(nctr):
            #idx = ops.index[(cs_of[i]+j*nprim):(cs_of[i]+(j+1)*nprim), ibas:(ibas+nbas), :]
            #grad = ops.index_add(grad, idx, g)
            grad[(cs_of[i]+j*nprim):(cs_of[i]+(j+1)*nprim), ibas:(ibas+nbas)] += g
            ibas += nbas
        off += nprim*nbas
    grad += grad.transpose(0,2,1)

    tangent_out = np.einsum('xij,x->ij', grad, ctr_coeff_t)
    #tangent_out += tangent_out.T
    return tangent_out

def _int1e_jvp_exp(mol, mol_t, intor):
    mol1 = _get_fakemol_exp(mol)
    mol1._atm[:,mole.CHARGE_OF] = 0 # set nuclear charge to zero
    if intor.endswith("_sph"):
        intor = intor.replace("_sph", "_cart")
        cart = False
    else:
        cart = True
        intor = mol._add_suffix(intor, cart=True)

    nbas1 = len(mol1._bas)
    nbas = len(mol._bas)
    shls_slice = (0, nbas1, nbas1, nbas1+nbas)
    if 'ECP' in intor:
        assert mol._ecp is not None
        bas = numpy.vstack((mol._bas, mol._ecpbas))
    else:
        bas = mol._bas
    atmc, basc, envc = mole.conc_env(mol1._atm, mol1._bas, mol1._env,
                                     mol._atm, bas, mol._env)
    if 'ECP' in intor:
        envc[mole.AS_ECPBAS_OFFSET] = len(mol1._bas) + len(mol._bas)
        envc[mole.AS_NECPBAS] = len(mol._ecpbas)

    s = moleintor.getints(intor, atmc, basc, envc, shls_slice)
    es, es_of, _env_of = setup_exp(mol)
    nao = mole.nao_cart(mol)
    #grad = np.zeros((len(es), nao, nao), dtype=float)
    grad = numpy.zeros((len(es), nao, nao), dtype=float)

    off = 0
    ibas = 0
    for i in range(len(mol._bas)):
        ioff = es_of[i]

        l = mol._bas[i,mole.ANG_OF]
        nbas = (l+1)*(l+2)//2
        nbas1 = (l+3)*(l+4)//2
        nprim = mol._bas[i,mole.NPRIM_OF]
        nctr = mol._bas[i,mole.NCTR_OF]
        ptr_ctr_coeff = mol._bas[i,mole.PTR_COEFF]
        g = s[off:off+nprim*nbas1].reshape(nprim, nbas1, nao)

        xyz = get_bas_label(l)
        xyz1 = get_bas_label(l+2)
        for k in range(nctr):
            for j in range(nprim):
                c = mol._env[ptr_ctr_coeff + k*nprim + j]
                if l == 0:
                    c *= 0.282094791773878143 # normalization factor for s orbital
                elif l == 1:
                    c *= 0.488602511902919921 # normalization factor for p orbital
                jbas = ibas
                for orb in xyz:
                    idx_x = xyz1.index(promote_xyz(orb, 'x', 2))
                    idx_y = xyz1.index(promote_xyz(orb, 'y', 2))
                    idx_z = xyz1.index(promote_xyz(orb, 'z', 2))
                    gc = -(g[j,idx_x] + g[j,idx_y] + g[j,idx_z]) *  c
                    #grad = ops.index_add(grad, ops.index[ioff+j, jbas], gc)
                    grad[ioff+j, jbas] += gc
                    jbas += 1
            ibas += nbas
        off += nprim * nbas1

    grad += grad.transpose(0,2,1)
    tangent_out = np.einsum('xij,x->ij', grad, mol_t.exp)
    #tangent_out += tangent_out.T
    if not mol.cart or not cart:
        c2s = np.asarray(mol.cart2sph_coeff())
        tangent_out = np.dot(c2s.T, np.dot(tangent_out, c2s))
    return tangent_out

def _int2e_jvp_r0(mol, mol_t, intor):
    coords_t = mol_t.coords
    #atmlst = range(mol.natm)
    #aoslices = numpy.asarray(mol.aoslice_by_atom(), dtype=numpy.int32)
    #nao = mol.nao

    #eri1 = -Mole.intor(mol, intor, comp=3, aosym='s2kl')
    #grad = numpy.zeros((mol.natm,3,nao,nao,nao,nao), dtype=numpy.double)
    #libcgto.restore_int2e_deriv(grad.ctypes.data_as(ctypes.c_void_p),
    #        eri1.ctypes.data_as(ctypes.c_void_p),
    #        aoslices.ctypes.data_as(ctypes.c_void_p),
    #        ctypes.c_int(mol.natm), ctypes.c_int(nao))
    #eri1 = None
    eri1 = -getints4c(mol, intor, comp=None, aosym='s1')
    aoslices = mol.aoslice_by_atom()[:,2:4]
    grad = _fill_grad_r0(eri1, aoslices)

    #for k, ia in enumerate(atmlst):
    #    p0, p1 = aoslices [ia,2:]
    #    tmp = np.einsum("xijkl,x->ijkl", eri1[:,p0:p1], coords_t[k])
    #    tangent_out = ops.index_add(tangent_out, ops.index[p0:p1], tmp)
    #    tangent_out = ops.index_add(tangent_out, ops.index[:,p0:p1], tmp.transpose((1,0,2,3)))
    #    tangent_out = ops.index_add(tangent_out, ops.index[:,:,p0:p1], tmp.transpose((2,3,0,1)))
    #    tangent_out = ops.index_add(tangent_out, ops.index[:,:,:,p0:p1], tmp.transpose((2,3,1,0)))
    return _int2e_dot_grad_tangent_r0(grad, coords_t)

@jit
def _int2e_dot_grad_tangent_r0(grad, tangent):
    tangent_out = np.einsum("nxijkl,nx->ijkl", grad, tangent)
    tangent_out += tangent_out.transpose(1,0,2,3)
    tangent_out += tangent_out.transpose(2,3,0,1)
    return tangent_out

def _gen_int2e_jvp_r0(mol, mol_t, intors):
    coords_t = mol_t.coords
    nao = mol.nao
    intor_a, intor_b, intor_c, intor_d = intors
    eri1_a = -getints4c(mol, intor_a, aosym='s1').reshape(3,-1,nao,nao,nao,nao)
    if intor_b:
        orders = _int2e_get_dr_order(intor_b)
        off = 3**orders[0]
        eri1_b = -getints4c(mol, intor_b, aosym='s1')
        eri1_b = eri1_b.reshape(off,3,-1,nao,nao,nao,nao)
        eri1_b = eri1_b.transpose(1,0,2,3,4,5,6).reshape(3,-1,nao,nao,nao,nao)
    else:
        eri1_b = eri1_a.transpose(0,1,3,2,4,5)

    orders = _int2e_get_dr_order(intor_c)
    off = 3**(orders[0]+orders[1])
    eri1_c = -getints4c(mol, intor_c, aosym='s1')
    eri1_c = eri1_c.reshape(off,3,-1,nao,nao,nao,nao)
    eri1_c = eri1_c.transpose(1,0,2,3,4,5,6).reshape(3,-1,nao,nao,nao,nao)
    if intor_d:
        orders = _int2e_get_dr_order(intor_d)
        off = 3**(orders[0]+orders[1]+orders[2])
        eri1_d = -getints4c(mol, intor_d, aosym='s1')
        eri1_d = eri1_d.reshape(off,3,-1,nao,nao,nao,nao)
        eri1_d = eri1_d.transpose(1,0,2,3,4,5,6).reshape(3,-1,nao,nao,nao,nao)
    else:
        eri1_d = eri1_c.transpose(0,1,2,3,5,4)

    aoslices = mol.aoslice_by_atom()[:,2:4]
    idx = np.arange(nao)
    idx_a = idx[None,None,:,None,None,None]
    idx_b = idx[None,None,None,:,None,None]
    idx_c = idx[None,None,None,None,:,None]
    idx_d = idx[None,None,None,None,None,:]
    grad = _gen_int2e_fill_grad_r0(eri1_a, eri1_b, eri1_c, eri1_d, aoslices,
                                   idx_a, idx_b, idx_c, idx_d)
    jvp = np.einsum("nxyijkl,nx->yijkl", grad, coords_t)
    return jvp

@jit
def _gen_int2e_fill_grad_r0(eri1_a, eri1_b, eri1_c, eri1_d, aoslices,
                            idx_a, idx_b, idx_c, idx_d):
    def body(slices):
        p0, p1 = slices[:]
        mask_a = (idx_a >= p0) & (idx_a < p1)
        mask_b = (idx_b >= p0) & (idx_b < p1)
        mask_c = (idx_c >= p0) & (idx_c < p1)
        mask_d = (idx_d >= p0) & (idx_d < p1)
        grad_a = np.where(mask_a, eri1_a, np.array(0, dtype=eri1_a.dtype))
        grad_b = np.where(mask_b, eri1_b, np.array(0, dtype=eri1_b.dtype))
        grad_c = np.where(mask_c, eri1_c, np.array(0, dtype=eri1_c.dtype))
        grad_d = np.where(mask_d, eri1_d, np.array(0, dtype=eri1_d.dtype))
        grad = grad_a + grad_b + grad_c + grad_d
        return grad
    grad = vmap(body)(aoslices)
    return grad

def _int2e_jvp_cs(mol, mol_t, intor):
    ctr_coeff = mol.ctr_coeff
    ctr_coeff_t = mol_t.ctr_coeff

    mol1 = _get_fakemol_cs(mol)

    nbas = len(mol._bas)
    nbas1 = len(mol1._bas)
    atmc, basc, envc = mole.conc_env(mol1._atm, mol1._bas, mol1._env,
                                     mol._atm, mol._bas, mol._env)

    shls_slice = (0, nbas1, nbas1, nbas1+nbas, nbas1, nbas1+nbas, nbas1, nbas1+nbas)
    intor = mol._add_suffix(intor)
    eri = moleintor.getints(intor, atmc, basc, envc, shls_slice)

    _, cs_of, _ = setup_ctr_coeff(mol)
    nao = mol.nao
    #grad = np.zeros((len(ctr_coeff), nao, nao, nao, nao), dtype=float)
    grad = numpy.zeros((len(ctr_coeff), nao, nao, nao, nao), dtype=eri.dtype)

    off = 0
    ibas = 0
    for i in range(len(mol._bas)):
        l = mol._bas[i,mole.ANG_OF]
        if mol.cart:
            nbas = (l+1)*(l+2)//2
        else:
            nbas = 2*l + 1
        nprim = mol._bas[i,mole.NPRIM_OF]
        nctr = mol._bas[i,mole.NCTR_OF]
        g = eri[off:(off+nprim*nbas)].reshape(nprim,-1,nao,nao,nao)
        for j in range(nctr):
            #idx = ops.index[(cs_of[i]+j*nprim):(cs_of[i]+(j+1)*nprim), ibas:(ibas+nbas)]
            #grad = ops.index_add(grad, idx, g)
            grad[(cs_of[i]+j*nprim):(cs_of[i]+(j+1)*nprim), ibas:(ibas+nbas)] += g
            ibas += nbas
        off += nprim*nbas
    #grad += grad.transpose(0,2,1,3,4)
    #grad += grad.transpose(0,3,4,1,2)
    tangent_out = _int2e_dot_grad_tangent_cs(grad, ctr_coeff_t)
    return tangent_out

@jit
def _int2e_dot_grad_tangent_cs(grad, tangent):
    tangent_out = np.einsum('xijkl,x->ijkl', grad, tangent)
    tangent_out += tangent_out.transpose(1,0,2,3)
    tangent_out += tangent_out.transpose(2,3,0,1)
    return tangent_out

def _int2e_jvp_exp(mol, mol_t, intor):
    mol1 = _get_fakemol_exp(mol)
    intor = mol._add_suffix(intor, cart=True)

    nbas = len(mol._bas)
    nbas1 = len(mol1._bas)
    atmc, basc, envc = mole.conc_env(mol1._atm, mol1._bas, mol1._env,
                                     mol._atm, mol._bas, mol._env)
    shls_slice = (0, nbas1, nbas1, nbas1+nbas, nbas1, nbas1+nbas, nbas1, nbas1+nbas)

    eri = moleintor.getints(intor, atmc, basc, envc, shls_slice)

    es, es_of, _env_of = setup_exp(mol)
    nao = mole.nao_cart(mol)
    #grad = np.zeros((len(es), nao, nao, nao, nao), dtype=float)
    grad = numpy.zeros((len(es), nao, nao, nao, nao), dtype=float)

    off = 0
    ibas = 0
    for i in range(len(mol._bas)):
        ioff = es_of[i]

        l = mol._bas[i,mole.ANG_OF]
        nbas = (l+1)*(l+2)//2
        nbas1 = (l+3)*(l+4)//2
        nprim = mol._bas[i,mole.NPRIM_OF]
        nctr = mol._bas[i,mole.NCTR_OF]
        ptr_ctr_coeff = mol._bas[i,mole.PTR_COEFF]
        g = eri[off:off+nprim*nbas1].reshape(nprim, nbas1, nao, nao, nao)

        xyz = get_bas_label(l)
        xyz1 = get_bas_label(l+2)
        for k in range(nctr):
            for j in range(nprim):
                c = mol._env[ptr_ctr_coeff + k*nprim + j]
                if l == 0:
                    c *= 0.282094791773878143 # normalization factor for s orbital
                elif l == 1:
                    c *= 0.488602511902919921 # normalization factor for p orbital
                jbas = ibas
                for orb in xyz:
                    idx_x = xyz1.index(promote_xyz(orb, 'x', 2))
                    idx_y = xyz1.index(promote_xyz(orb, 'y', 2))
                    idx_z = xyz1.index(promote_xyz(orb, 'z', 2))
                    gc = -(g[j,idx_x] + g[j,idx_y] + g[j,idx_z]) *  c
                    #grad = ops.index_add(grad, ops.index[ioff+j, jbas], gc)
                    grad[ioff+j, jbas] += gc
                    jbas += 1
            ibas += nbas
        off += nprim * nbas1

    #grad += grad.transpose(0,2,1,3,4)
    #grad += grad.transpose(0,3,4,1,2)
    tangent_out = _int2e_dot_grad_tangent_exp(grad, mol_t.exp)
    if not mol.cart:
        c2s = numpy.asarray(mol.cart2sph_coeff())
        tangent_out = _int2e_c2s(tangent_out, c2s)
    return tangent_out

@jit
def _int2e_dot_grad_tangent_exp(grad, tangent):
    tangent_out = np.einsum('xijkl,x->ijkl', grad, tangent)
    tangent_out += tangent_out.transpose(1,0,2,3)
    tangent_out += tangent_out.transpose(2,3,0,1)
    return tangent_out

@jit
def _int2e_c2s(eris_cart, c2s):
    eris_sph = np.einsum("iu,jv,ijkl,ks,lt->uvst", c2s, c2s, eris_cart, c2s, c2s)
    return eris_sph



from functools import reduce
from pyscf import numpy as np
#from pyscfad.lib import numpy as np
from pyscfad import ao2mo
from pyscfad.fci import fci_slow
from pyscfad.fci.fci_slow import fci_ovlp

def solve_fci(mf, nroots=1):
    mol = mf.mol
    mo_coeff = mf.mo_coeff
    norb = mo_coeff.shape[-1]
    nelec = mol.nelectron
    h1e = reduce(np.dot, (mo_coeff.T, mf.get_hcore(), mo_coeff))
    eri = ao2mo.incore.full(mf._eri, mo_coeff, compact=False)
    eri = eri.reshape(norb,norb,norb,norb)
    e, fcivec = fci_slow.kernel(h1e, eri, norb, nelec, mf.energy_nuc(), nroots=nroots)
    return e, fcivec



from functools import partial
import scipy.linalg
from pyscf import numpy as np
from pyscfad.lib import ops, custom_jvp, jit, linalg_helper

DEG_THRESH = 1e-10

def eigh(a, b=None, x0=None, eigvals_only=False, **kwargs):
    if x0 is None:
        return linalg_helper.eigh(a, b, eigvals_only=eigvals_only, **kwargs)

    a = 0.5 * (a + a.T.conj())
    if b is not None:
        b = 0.5 * (b + b.T.conj())

    w, v =  _eigh(a, b, x0)

    if eigvals_only:
        return w
    else:
        return w, v

@partial(custom_jvp, nondiff_argnums=(2,))
def _eigh(a, b, x0):
    w, v = scipy.linalg.eigh(a, b=b)
    w = np.asarray(w, dtype=v.dtype)
    return w, v

@_eigh.defjvp
def _eigh_jvp(v, primals, tangents):
    a, b = primals
    at, bt = tangents
    w, _ = primal_out = _eigh(a, b, v)

    deg_thresh = DEG_THRESH
    eji = w[..., np.newaxis, :] - w[..., np.newaxis]
    idx = abs(eji) < deg_thresh
    #eji[idx] = 1.e200
    #eji[numpy.diag_indices_from(eji)] = 1
    eji = ops.index_update(eji, idx, 1.e200)
    eji = ops.index_update(eji, np.diag_indices_from(eji), 1.)
    eye_n = np.eye(a.shape[-1], dtype=a.dtype)
    Fmat = np.reciprocal(eji) - eye_n
    if b is None:
        dw, dv = _eigh_jvp_jitted_nob(v, Fmat, at)
    else:
        dw, dv = _eigh_jvp_jitted(w, v, Fmat, at, bt)
        if idx.sum() > 0:
            mask = np.zeros_like(b, dtype=np.int32)
            mask = mask.at[idx].set(1)
            mask -= eye_n
            tmp = -.5 * np.dot(v.conj().T, np.dot(bt, v))
            dv += np.dot(v, np.multiply(mask, tmp))
    return primal_out, (dw,dv)

@jit
def _eigh_jvp_jitted(w, v, Fmat, at, bt):
    vt_at_v = np.dot(v.conj().T, np.dot(at, v))
    vt_bt_v = np.dot(v.conj().T, np.dot(bt, v))
    vt_bt_v_w = np.dot(vt_bt_v, np.diag(w))
    da_minus_ds = vt_at_v - vt_bt_v_w
    dw = np.diag(da_minus_ds)#.real

    eye_n = np.eye(vt_bt_v.shape[-1])
    dv = np.dot(v, np.multiply(Fmat, da_minus_ds) - np.multiply(eye_n, vt_bt_v) * .5)
    return dw, dv

@jit
def _eigh_jvp_jitted_nob(v, Fmat, at):
    vt_at_v = np.dot(v.conj().T, np.dot(at, v))
    dw = np.diag(vt_at_v)
    dv = np.dot(v, np.multiply(Fmat, vt_at_v))
    return dw, dv



from functools import partial
import numpy
from jax import custom_jvp
from pyscf.scf import _vhf

@partial(custom_jvp, nondiff_argnums=(2,3,4))
def incore(eri, dms, hermi=0, with_j=True, with_k=True):
    eri = numpy.asarray(eri)
    dms = numpy.asarray(dms)
    return _vhf.incore(eri, dms, hermi, with_j, with_k)

@incore.defjvp
def incore_jvp(hermi, with_j, with_k,
               primals, tangents):
    raise NotImplementedError
    #eri, dms, = primals
    #eri_t, dms_t, = tangents

    #vj, vk = incore(eri, dms, hermi, with_j, with_k)

    #nao = dms_shape[-1]
    #npair = nao*(nao+1)//2
    #if eri.ndim == 2 and npair*npair == eri.size:
    #    pass
    #elif eri.ndim == 1 and npair*(npair+1)//2 == eri.size:
    #    pass
    #else:
    #    raise RuntimeError

    #vj_dot = vk_dot = None
    #return (vj, vk), (vj_dot, vk_dot)



from functools import reduce
from jax import vmap
from pyscf import numpy as np
from pyscf.scf import diis as pyscf_cdiis
from pyscfad.lib import jit
from pyscfad.lib import diis

class CDIIS(diis.DIIS, pyscf_cdiis.CDIIS):
    def __init__(self, mf=None, filename=None):
        pyscf_cdiis.CDIIS.__init__(self, mf=mf, filename=filename)
        self.incore = True

@jit
def get_err_vec(s, d, f):
    '''error vector = SDF - FDS'''
    f = np.asarray(f)
    s = np.asarray(s).reshape(f.shape)
    d = np.asarray(d).reshape(f.shape)

    def body(s, d, f):
        sdf = reduce(np.dot, (s,d,f))
        return sdf.T.conj() - sdf

    if f.ndim == 2:
        errvec = body(s, d, f)
    else:
        errvec = vmap(body)(s, d, f)
    return errvec

SCFDIIS = SCF_DIIS = DIIS = CDIIS



from functools import partial
import numpy
import scipy
import scipy.linalg
from pyscf import numpy as np
from pyscfad.lib import ops, custom_jvp, jit

# threshold for degenerate eigenvalues
DEG_THRESH = 1e-10

def eigh(a, b=None, lower=True, eigvals_only=False, overwrite_a=False,
         overwrite_b=False, turbo=True, eigvals=None, type=1,
         check_finite=True, subset_by_index=None, subset_by_value=None,
         driver=None):
    if overwrite_a is True or overwrite_b is True:
        raise NotImplementedError("Overwritting a or b is not implemeneted.")
    if type != 1:
        raise NotImplementedError("Only the type=1 case of eigh is implemented.")
    if not(eigvals is None and subset_by_index is None and subset_by_value is None):
        raise NotImplementedError("Subset of eigen values is not implemented.")

    a = 0.5 * (a + a.T.conj())
    if b is not None:
        b = 0.5 * (b + b.T.conj())

    w, v =  _eigh(a, b, lower=lower,
                  turbo=turbo, check_finite=check_finite, driver=driver)

    if eigvals_only:
        return w
    else:
        return w, v

@partial(custom_jvp, nondiff_argnums=(2,3,4,5))
def _eigh(a, b, lower=True,
          turbo=True, check_finite=True, driver=None):
    if scipy.__version__ >= "1.5.0":
        w, v = scipy.linalg.eigh(a, b=b, lower=lower, check_finite=check_finite, driver=driver)
    else:
        w, v = scipy.linalg.eigh(a, b=b, lower=lower, check_finite=check_finite, turbo=turbo)
    w = np.asarray(w, dtype=v.dtype)
    return w, v

@_eigh.defjvp
def _eigh_jvp(lower, turbo, check_finite, driver, primals, tangents):
    a, b = primals
    at, bt = tangents
    w, v = primal_out = _eigh(*primals,
                              lower=lower, turbo=turbo, check_finite=check_finite, driver=driver)

    deg_thresh = DEG_THRESH
    eji = w[..., numpy.newaxis, :] - w[..., numpy.newaxis]
    idx = abs(eji) < deg_thresh
    #eji[idx] = 1.e200
    #eji[numpy.diag_indices_from(eji)] = 1
    eji = ops.index_update(eji, idx, 1.e200)
    eji = ops.index_update(eji, np.diag_indices_from(eji), 1.)
    eye_n = numpy.eye(a.shape[-1], dtype=a.dtype)
    Fmat = np.reciprocal(eji) - eye_n
    if b is None:
        dw, dv = _eigh_jvp_jitted_nob(v, Fmat, at)
    else:
        dw, dv = _eigh_jvp_jitted(w, v, Fmat, at, bt)
        if idx.sum() > 0:
            mask = np.zeros_like(b, dtype=np.int32)
            mask = mask.at[idx].set(1)
            mask -= eye_n
            tmp = -.5 * np.dot(v.conj().T, np.dot(bt, v))
            dv += np.dot(v, np.multiply(mask, tmp))
    return primal_out, (dw,dv)

@jit
def _eigh_jvp_jitted(w, v, Fmat, at, bt):
    vt_at_v = np.dot(v.conj().T, np.dot(at, v))
    vt_bt_v = np.dot(v.conj().T, np.dot(bt, v))
    vt_bt_v_w = np.dot(vt_bt_v, np.diag(w))
    da_minus_ds = vt_at_v - vt_bt_v_w
    dw = np.diag(da_minus_ds)#.real

    eye_n = np.eye(vt_bt_v.shape[-1])
    dv = np.dot(v, np.multiply(Fmat, da_minus_ds) - np.multiply(eye_n, vt_bt_v) * .5)
    return dw, dv

@jit
def _eigh_jvp_jitted_nob(v, Fmat, at):
    vt_at_v = np.dot(v.conj().T, np.dot(at, v))
    dw = np.diag(vt_at_v)
    dv = np.dot(v, np.multiply(Fmat, vt_at_v))
    return dw, dv



import numpy
from pyscf import __config__
from pyscf.lib import logger
from pyscf.pbc.dft import rks as pyscf_rks
from pyscf.pbc.dft import gen_grid, multigrid
from pyscf.pbc.dft.rks import prune_small_rho_grids_
from pyscfad import util
from pyscfad.lib import numpy as np
from pyscfad.lib import stop_grad
from pyscfad.dft import rks as mol_ks
from pyscfad.dft.rks import VXC
from pyscfad.pbc.scf import hf as pbchf
from pyscfad.pbc.dft import numint

def get_veff(ks, cell=None, dm=None, dm_last=0, vhf_last=0, hermi=1,
             kpt=None, kpts_band=None):
    if cell is None:
        cell = ks.cell
    if dm is None:
        dm = ks.make_rdm1()
    if kpt is None:
        kpt = ks.kpt

    log = logger.new_logger(ks)

    omega, alpha, hyb = ks._numint.rsh_and_hybrid_coeff(ks.xc, spin=cell.spin)
    hybrid = abs(hyb) > 1e-10 or abs(alpha) > 1e-10

    if not hybrid and isinstance(ks.with_df, multigrid.MultiGridFFTDF):
        n, exc, vxc = multigrid.nr_rks(ks.with_df, ks.xc, dm, hermi,
                                       kpt.reshape(1,3), kpts_band,
                                       with_j=True, return_j=False)
        log.debug('nelec by numeric integration = %s', n)
        log.timer('vxc')
        return vxc

    ground_state = (getattr(dm, 'ndim', 0) == 2 and kpts_band is None)

    if ks.grids.non0tab is None:
        ks.grids.build(with_non0tab=True)
        if (isinstance(ks.grids, gen_grid.BeckeGrids) and
            ks.small_rho_cutoff > 1e-20 and ground_state):
            ks.grids = prune_small_rho_grids_(ks, stop_grad(cell), stop_grad(dm), ks.grids, kpt)
        log.timer('setting up grids')

    if hermi == 2:  # because rho = 0
        n, exc, vxc = 0, 0, 0
    else:
        n, exc, vxc = ks._numint.nr_rks(cell, ks.grids, ks.xc, dm, 0,
                                        kpt, kpts_band)
        log.debug('nelec by numeric integration = %s', n)
        log.timer('vxc')

    if not hybrid:
        vj = ks.get_j(cell, dm, hermi, kpt, kpts_band)
        vxc += vj
    else:
        if getattr(ks.with_df, '_j_only', False):  # for GDF and MDF
            ks.with_df._j_only = False
        vj, vk = ks.get_jk(cell, dm, hermi, kpt, kpts_band)
        vk *= hyb
        if abs(omega) > 1e-10:
            vklr = ks.get_k(cell, dm, hermi, kpt, kpts_band, omega=omega)
            vklr *= (alpha - hyb)
            vk += vklr
        vxc += vj - vk * .5

        if ground_state:
            exc -= np.einsum('ij,ji', dm, vk).real * .5 * .5

    if ground_state:
        ecoul = np.einsum('ij,ji', dm, vj).real * .5
    else:
        ecoul = None

    vxc = VXC(vxc=vxc, ecoul=ecoul, exc=exc, vj=None, vk=None)
    del log
    return vxc

def _dft_common_init_(mf, xc='LDA,VWN', **kwargs):
    from pyscfad.pbc.scf import khf
    mf.xc = xc
    mf.grids = None
    mf.small_rho_cutoff = getattr(__config__, 'dft_rks_RKS_small_rho_cutoff', 1e-7)
    if isinstance(mf, khf.KSCF):
        mf._numint = numint.KNumInt(mf.kpts)
    else:
        mf._numint = numint.NumInt()
    mf._keys = mf._keys.union(['xc', 'grids', 'small_rho_cutoff'])

def _dft_common_post_init_(mf):
    if mf.grids is None:
        mf.grids = gen_grid.UniformGrids(stop_grad(mf.cell))

class KohnShamDFT(mol_ks.KohnShamDFT):
    __init__ = _dft_common_init_
    __post_init__ = _dft_common_post_init_

@util.pytree_node(pbchf.Traced_Attributes, num_args=1)
class RKS(KohnShamDFT, pbchf.RHF):
    def __init__(self, cell, xc='LDA,VWN', kpt=numpy.zeros(3),
                 exxdiv=getattr(__config__, 'pbc_scf_SCF_exxdiv', 'ewald'), **kwargs):
        pbchf.RHF.__init__(self, cell, kpt, exxdiv, **kwargs)
        KohnShamDFT.__init__(self, xc)
        self.__dict__.update(kwargs)
        # NOTE this has to be after __dict__ update,
        # otherwise stop_grad(mol) won't work.
        # Currently, no grid response is considered.
        KohnShamDFT.__post_init__(self)

    def dump_flags(self, verbose=None):
        pbchf.RHF.dump_flags(self, verbose)
        KohnShamDFT.dump_flags(self, verbose)
        return self

    get_veff = get_veff
    energy_elec = mol_ks.energy_elec
    get_rho = pyscf_rks.get_rho



from pyscf import numpy as np
from pyscf.lib import logger
from pyscfad.lib import jit, vmap

'''
CCSD(T)
'''

# t3 as ijkabc

# JCP 94, 442 (1991); DOI:10.1063/1.460359.  Error in Eq (1), should be [ia] >= [jb] >= [kc]
def kernel(mycc, eris, t1=None, t2=None, verbose=logger.NOTE):
    log = logger.new_logger(mycc, verbose)

    if t1 is None: t1 = mycc.t1
    if t2 is None: t2 = mycc.t2

    t1T = t1.T
    t2T = t2.transpose(2,3,0,1)

    nocc, nvir = t1.shape
    mo_e = eris.mo_energy
    e_occ, e_vir = mo_e[:nocc], mo_e[nocc:]
    #eijk = lib.direct_sum('i,j,k->ijk', e_occ, e_occ, e_occ)
    eijk = e_occ[:,None,None] + e_occ[None,:,None] + e_occ[None,None,:]

    eris_vvov = eris.get_ovvv().conj().transpose(1,3,0,2)
    eris_vooo = np.asarray(eris.ovoo).conj().transpose(1,0,3,2)
    eris_vvoo = np.asarray(eris.ovov).conj().transpose(1,3,0,2)
    fvo = eris.fock[nocc:,:nocc]

    idx = []
    scal = []
    for a in range(nvir):
        for b in range(a+1):
            for c in range(b+1):
                idx.append([a,b,c])
                if a == c:
                    fac = 6
                elif a == b or b == c:
                    fac = 2
                else:
                    fac = 1
                scal.append(fac)

    idx = np.asarray(idx, dtype=np.int32)
    scal = np.asarray(scal, dtype=float)

    et = _compute_et(t1, t2, eris_vvov, eris_vooo, eris_vvoo,
                     fvo, eijk, e_vir, idx, scal)
    log.info('CCSD(T) correction = %.15g', et)
    return et

@jit
def _compute_et(t1, t2, eris_vvov, eris_vooo, eris_vvoo,
                fvo, eijk, e_vir, idx, scal):
    t1T = t1.T
    t2T = t2.transpose(2,3,0,1)

    def get_w(a, b, c):
        w = np.einsum('if,fkj->ijk', eris_vvov[a,b], t2T[c,:])
        w-= np.einsum('ijm,mk->ijk', eris_vooo[a,:], t2T[b,c])
        return w

    def get_v(a, b, c):
        v = np.einsum('ij,k->ijk', eris_vvoo[a,b], t1T[c])
        v+= np.einsum('ij,k->ijk', t2T[a,b], fvo[c])
        return v

    def r3(w):
        return (4 * w + w.transpose(1,2,0) + w.transpose(2,0,1)
                - 2 * w.transpose(2,1,0) - 2 * w.transpose(0,2,1)
                - 2 * w.transpose(1,0,2))

    def body(fac, index):
        a, b, c = index[:]
        d3  = eijk - e_vir[a] - e_vir[b] - e_vir[c]
        d3 *= fac

        wabc = get_w(a, b, c)
        wacb = get_w(a, c, b)
        wbac = get_w(b, a, c)
        wbca = get_w(b, c, a)
        wcab = get_w(c, a, b)
        wcba = get_w(c, b, a)
        vabc = get_v(a, b, c)
        vacb = get_v(a, c, b)
        vbac = get_v(b, a, c)
        vbca = get_v(b, c, a)
        vcab = get_v(c, a, b)
        vcba = get_v(c, b, a)
        zabc = r3(wabc + .5 * vabc) / d3
        zacb = r3(wacb + .5 * vacb) / d3
        zbac = r3(wbac + .5 * vbac) / d3
        zbca = r3(wbca + .5 * vbca) / d3
        zcab = r3(wcab + .5 * vcab) / d3
        zcba = r3(wcba + .5 * vcba) / d3

        et = np.einsum('ijk,ijk', wabc, zabc.conj())
        et+= np.einsum('ikj,ijk', wacb, zabc.conj())
        et+= np.einsum('jik,ijk', wbac, zabc.conj())
        et+= np.einsum('jki,ijk', wbca, zabc.conj())
        et+= np.einsum('kij,ijk', wcab, zabc.conj())
        et+= np.einsum('kji,ijk', wcba, zabc.conj())

        et+= np.einsum('ijk,ijk', wacb, zacb.conj())
        et+= np.einsum('ikj,ijk', wabc, zacb.conj())
        et+= np.einsum('jik,ijk', wcab, zacb.conj())
        et+= np.einsum('jki,ijk', wcba, zacb.conj())
        et+= np.einsum('kij,ijk', wbac, zacb.conj())
        et+= np.einsum('kji,ijk', wbca, zacb.conj())

        et+= np.einsum('ijk,ijk', wbac, zbac.conj())
        et+= np.einsum('ikj,ijk', wbca, zbac.conj())
        et+= np.einsum('jik,ijk', wabc, zbac.conj())
        et+= np.einsum('jki,ijk', wacb, zbac.conj())
        et+= np.einsum('kij,ijk', wcba, zbac.conj())
        et+= np.einsum('kji,ijk', wcab, zbac.conj())

        et+= np.einsum('ijk,ijk', wbca, zbca.conj())
        et+= np.einsum('ikj,ijk', wbac, zbca.conj())
        et+= np.einsum('jik,ijk', wcba, zbca.conj())
        et+= np.einsum('jki,ijk', wcab, zbca.conj())
        et+= np.einsum('kij,ijk', wabc, zbca.conj())
        et+= np.einsum('kji,ijk', wacb, zbca.conj())

        et+= np.einsum('ijk,ijk', wcab, zcab.conj())
        et+= np.einsum('ikj,ijk', wcba, zcab.conj())
        et+= np.einsum('jik,ijk', wacb, zcab.conj())
        et+= np.einsum('jki,ijk', wabc, zcab.conj())
        et+= np.einsum('kij,ijk', wbca, zcab.conj())
        et+= np.einsum('kji,ijk', wbac, zcab.conj())

        et+= np.einsum('ijk,ijk', wcba, zcba.conj())
        et+= np.einsum('ikj,ijk', wcab, zcba.conj())
        et+= np.einsum('jik,ijk', wbca, zcba.conj())
        et+= np.einsum('jki,ijk', wbac, zcba.conj())
        et+= np.einsum('kij,ijk', wacb, zcba.conj())
        et+= np.einsum('kji,ijk', wabc, zcba.conj())
        return et

    et = vmap(body, in_axes=(0,0), signature='(),(x)->()')(scal, idx)
    et = np.sum(et) * 2
    return et



from pyscf import numpy as np
from pyscf.fci import cistring
from pyscf.fci import fci_slow as pyscf_fci_slow
#from pyscfad.lib import numpy as np
from pyscfad.lib import vmap
from pyscfad.gto import mole

def get_occ_loc(strs, norb):
    locs = []
    for x in strs:
        loc = []
        for i in range(norb):
            if (x >> i) & 1 == 1:
                loc.append(i)
        locs.append(loc)
    return np.asarray(locs)

def fci_ovlp(mol1, mol2, fcivec1, fcivec2, norb1, norb2, nelec1, nelec2, mo1, mo2):
    mo1 = np.asarray(mo1)
    mo2 = np.asarray(mo2)
    fcivec1 = np.asarray(fcivec1)
    fcivec2 = np.asarray(fcivec2)
    if isinstance(nelec1, (int, np.integer)):
        nelecb1 = nelec1//2
        neleca1 = nelec1 - nelecb1
    else:
        neleca1, nelecb1 = nelec1

    if isinstance(nelec2, (int, np.integer)):
        nelecb2 = nelec2//2
        neleca2 = nelec2 - nelecb2
    else:
        neleca2, nelecb2 = nelec2

    sao = mole.intor_cross('int1e_ovlp', mol1, mol2)

    strs_a1 = cistring.make_strings(range(norb1), neleca1)
    locs_a1 = get_occ_loc(strs_a1, norb1)
    if neleca1 == nelecb1:
        locs_b1 = locs_a1
    else:
        strs_b1 = cistring.make_strings(range(norb1), nelecb1)
        locs_b1 = get_occ_loc(strs_b1, norb1)

    if norb1 == norb2 and neleca1 == neleca2:
        locs_a2 = locs_a1
    else:
        strs_a2 = cistring.make_strings(range(norb2), neleca2)
        locs_a2 = get_occ_loc(strs_a2, norb2)
    if norb1 == norb2 and nelecb1 == nelecb2:
        locs_b2 = locs_b1
    else:
        strs_b2 = cistring.make_strings(range(norb2), nelecb2)
        locs_b2 = get_occ_loc(strs_b2, norb2)

    na1 = len(locs_a1)
    nb1 = len(locs_b1)
    na2 = len(locs_a2)
    nb2 = len(locs_b2)

    if getattr(mo1, 'ndim', None) == 2:
        mo_a1 = mo_b1 = mo1
    else:
        mo_a1 = mo1[0]
        mo_b1 = mo1[1]
    if getattr(mo2, 'ndim', None) == 2:
        mo_a2 = mo_b2 = mo2
    else:
        mo_a2 = mo2[0]
        mo_b2 = mo2[1]

    ci1 = fcivec1.reshape(na1,nb1)
    ci2 = fcivec2.reshape(na2,nb2)

    idxa = np.broadcast_to(locs_a2[:,None,:], (na2,nb2,neleca2)).reshape(-1,neleca2)
    idxb = np.broadcast_to(locs_b2[None,:,:], (na2,nb2,nelecb2)).reshape(-1,nelecb2)
    def body(mo_ia, mo_ib, ida, idb):
        sij_a = np.einsum('ui,uv,vj->ij', mo_ia, sao, mo_a2[:,ida])
        sij_b = np.einsum('ui,uv,vj->ij', mo_ib, sao, mo_b2[:,idb])
        val = np.linalg.det(sij_a) * np.linalg.det(sij_b)
        return val

    res = 0.
    for ia in range(na1):
        mo_ia = mo_a1[:,locs_a1[ia]]
        for ib in range(nb1):
            mo_ib = mo_b1[:,locs_b1[ib]]
            val = vmap(body, (None,None,0,0), signature='(i),(j)->()')(mo_ia, mo_ib, idxa, idxb)
            #val = []
            #for i in range(len(idxa)):
            #    val.append(body(mo_ia, mo_ib, idxa[i], idxb[i]))
            #val = np.asarray(val)
            res += ci1[ia,ib] * (val * ci2.ravel()).sum()
    return res


kernel = pyscf_fci_slow.kernel
